"""
Network connector for Curve Prices API.
"""

from typing import List

from eth_utils import to_checksum_address
from pandas import DataFrame, to_datetime

from curvesim.exceptions import ApiResultError, CurvesimValueError

from .http import HTTP
from .utils import sync

URL = "https://prices.curve.fi/v1/"

CHAIN_ALIASES = {"mainnet": "ethereum"}


async def _get_pool_pair_volume(
    pool_address,
    main_token_address,
    reference_token_address,
    start_ts,
    end_ts,
    *,
    chain="ethereum",
    interval="day",
):
    chain = _chain_from_alias(chain)
    pool_address = to_checksum_address(pool_address)
    main_token_address = to_checksum_address(main_token_address)
    reference_token_address = to_checksum_address(reference_token_address)

    url = URL + f"volume/{chain}/{pool_address}"
    params = {
        "main_token": main_token_address,
        "reference_token": reference_token_address,
        "start": start_ts,
        "end": end_ts,
        "interval": interval,
    }
    r = await HTTP.get(url, params=params)

    try:
        data = r["data"]
    except KeyError as e:
        raise ApiResultError(
            "No historical volume returned for\n"
            f"Pool: '{pool_address}', Chain: '{chain}',\n"
            f"Tokens: (main: {main_token_address}, "
            f"reference: {reference_token_address}),\n"
            f"Timestamps: (start: {start_ts}, end: {end_ts})"
        ) from e

    return data


async def get_pool_pair_volume(
    pool_address: str,
    main_token_address: str,
    reference_token_address: str,
    start_ts: int,
    end_ts: int,
    *,
    chain: str = "ethereum",
    interval: str = "day",
) -> DataFrame:
    """
    Gets historical daily volume for a pair of coins traded in a Curve pool.

    Parameters
    ----------
    pool_address: str
        The Curve pool's address.

    main_token_address: str
        Address for the token volume will be denominated in.

    reference_token_address: str
        Address for the second token in the trading pair.

    start_ts: int
        Posix timestamp (UTC) for start of query period.

    end_ts: int
        Posix timestamp (UTC) for end of query period.

    chain: str, default "ethereum"
        The pool's blockchain (note: currently only "ethereum" supported)

    interval: str, default "day"
        The sampling interval for the data. Available values: week, day, hour

    Returns
    -------
    DataFrame
        Rows: DateTimeIndex; Columns: volume, fees

    """
    data: List[dict] = await _get_pool_pair_volume(
        pool_address,
        main_token_address,
        reference_token_address,
        start_ts,
        end_ts,
        chain=chain,
        interval=interval,
    )

    df = DataFrame(data, columns=["timestamp", "volume", "fees"])
    df["timestamp"] = to_datetime(df["timestamp"], unit="s")
    df.set_index("timestamp", inplace=True)
    return df


def _chain_from_alias(chain):
    if chain in CHAIN_ALIASES:  # pylint: disable=consider-using-get
        chain = CHAIN_ALIASES[chain]

    if chain != "ethereum":
        raise CurvesimValueError(
            "Curve Prices API currently only supports Ethereum chain."
        )

    return chain


get_pool_pair_volume_sync = sync(get_pool_pair_volume)



"""
Mainly a module to house the `CurveCryptoPool`, a cryptoswap implementation in Python.
"""
import time
from math import isqrt, prod
from typing import List, Optional, Tuple, Type

from curvesim.exceptions import CalculationError, CryptoPoolError, CurvesimValueError
from curvesim.logging import get_logger
from curvesim.pool.base import Pool
from curvesim.pool.snapshot import CurveCryptoPoolBalanceSnapshot, Snapshot

from .calcs import (
    factory_2_coin,
    geometric_mean,
    get_alpha,
    get_p,
    get_y,
    newton_D,
    tricrypto_ng,
)

logger = get_logger(__name__)

NOISE_FEE = 10**5  # 0.1 bps
EXP_PRECISION = 10**10
PRECISION = 10**18


class CurveCryptoPool(Pool):  # pylint: disable=too-many-instance-attributes
    """Cryptoswap implementation in Python."""

    snapshot_class: Type[Snapshot] = CurveCryptoPoolBalanceSnapshot

    __slots__ = (
        "A",
        "gamma",
        "n",
        "precisions",
        "mid_fee",
        "out_fee",
        "allowed_extra_profit",
        "fee_gamma",
        "adjustment_step",
        "admin_fee",
        "ma_half_time",
        "price_scale",
        "_price_oracle",
        "last_prices",
        "last_prices_timestamp",
        "_block_timestamp",
        "balances",
        "D",
        "virtual_price",
        "tokens",
        "xcp_profit",
        "xcp_profit_a",
        "not_adjusted",
    )

    # pylint: disable-next=too-many-locals,too-many-arguments,too-many-branches
    def __init__(
        self,
        A: int,
        gamma: int,
        n: int,
        precisions: List[int],
        mid_fee: int,
        out_fee: int,
        allowed_extra_profit: int,
        fee_gamma: int,
        adjustment_step: int,
        ma_half_time: int,
        price_scale: List[int],
        price_oracle=None,
        last_prices=None,
        last_prices_timestamp=None,
        balances=None,
        D=None,
        tokens=None,
        admin_fee: int = 5 * 10**9,
        xcp_profit=10**18,
        xcp_profit_a=10**18,
        virtual_price=None,
    ) -> None:
        """
        Parameters
        ----------
        A : int
            Amplification coefficient; this is :math:`A n^n` in the whitepaper
            multiplied by 10**4 for greater precision.
        gamma: int
            Decay factor for A.
        n: int
            Number of coins; currently only n = 2 is supported.
        precisions: list of int
            Precision adjustments to convert native token units to 18 decimals;
            this assumes tokens have at most 18 decimals
            i.e. balance in native units * precision = balance in D units
        mid_fee: int
            Fee with 10**10 precision, for trades near price scale
        out_fee: int
            Fee with 10**10 precision, used to adjust `mid_fee` for trades
            further away from price_scale
        allowed_extra_profit: int
            "Buffer" used to determine if the price adjustment algorithm
            should run.
        fee_gamma: int
            Factor used to control the transition from `mid_fee` to `out_fee`.
        adjustment_step:
            Minimum step size to adjust the price scale.
        ma_half_time: int
            "Half-life" for exponential moving average of trade prices.
        price_scale: List[Int]
            Price scale value for the pool.  This is where liquidity is concentrated.
        price_oracle: List[Int], optional
            Price oracle value for the pool.  This is the EMA price used to
            adjust the price scale toward.
            Defaults to `price_scale`.
        last_prices: List[Int], optional
            Last trade price for the pool.
            Defaults to `price_scale`.
        last_prices_timestamp: int, optional
            Timestamp for last operation altering pool price.
            Defaults to unix timestamp.
        balances: list of int, optional
            Coin balances in native token units;
            either `balances` or `D` is required
        D : int, optional
            Stableswap invariant for given balances, precisions, price_scale,
            A, and gamma; either `balances` or `D` is required
        tokens: int, optional
            LP token supply (default is calculated from `D`, which is also
            calculated if needed)
        admin_fee: int, optional
            Percentage of `fee` with 10**10 precision.  Fee paid to the DAO
            (default = 5*10**9)
        xcp_profit: int, optional
            Counter for accumulated profits, no losses (default = 10**18)
        xcp_profit_a: int, optional
            Value of `xcp_profit` when admin fees last claimed (default = 10**18)
        virtual_price: int, optional
            amount of XCP invariant per LP token; can be used when
            missing `tokens` value.
        """
        self.A = A
        self.gamma = gamma

        self.mid_fee = mid_fee
        self.out_fee = out_fee
        self.allowed_extra_profit = allowed_extra_profit
        self.fee_gamma = fee_gamma
        self.adjustment_step = adjustment_step
        self.admin_fee = admin_fee

        self.price_scale = price_scale.copy()
        self._price_oracle = price_oracle.copy() if price_oracle else price_scale.copy()
        self.last_prices = last_prices.copy() if last_prices else price_scale.copy()
        self.ma_half_time = ma_half_time

        self._block_timestamp = _get_unix_timestamp()
        self.last_prices_timestamp = last_prices_timestamp or 0

        self.xcp_profit = xcp_profit
        self.xcp_profit_a = xcp_profit_a  # Full profit at last claim of admin fees
        self.not_adjusted = False

        if n not in [2, 3]:
            raise CryptoPoolError(
                "Only 2 or 3-coin crypto pools are currently supported."
            )
        self.n = n
        self.precisions = precisions

        if len(precisions) != n:
            raise ValueError("`len(precisions)` must equal `n`")

        if balances is None and D is None:
            raise ValueError("Must provide at least one of `balances` or `D`.")

        # All state variables needed for balance conversions or "newton"
        # calculations should have been set by this point.

        if balances:
            self.balances = balances.copy()

        if D is not None:
            self.D = D
            if not balances:
                self.balances = self._convert_D_to_balances(D)
            else:
                # If user passes both `D` and `balances`, it's possible they may
                # be inconsistent; however we allow this for unanticipated use-cases.
                logger.debug(
                    "Both `D` and `balances` were passed into `__init__`. "
                    "Inconsistent values may create issues."
                )
        else:
            xp = self._xp()
            D = newton_D(A, gamma, xp)
            self.D = D

        if tokens and virtual_price:
            raise CurvesimValueError(
                "Should not set both `tokens` and `virtual_price`."
            )

        xcp = self._get_xcp(D)
        if virtual_price:
            self.virtual_price = virtual_price
            self.tokens = xcp * 10**18 // virtual_price
        else:
            if tokens:
                self.tokens = tokens
            else:
                self.tokens = xcp
            self.virtual_price = 10**18 * xcp // self.tokens

    def _convert_D_to_balances(self, D):
        price_scale = self.price_scale
        precisions = self.precisions
        n = self.n

        return [D // n // precisions[0]] + [
            D * PRECISION // (p * n) // prec
            for p, prec in zip(price_scale, precisions[1:])
        ]

    def _xp(self) -> List[int]:
        """
        Calculate the balances in units of `D`, converting using `price_scale`
        so a unit of each token has equal value.

        Returns
        --------
        List[int]
            The pool balances in units of `D`.

        Note
        -----
        This intentionally always return a new copy of the balances.
        """
        balances = self.balances
        return self._xp_mem(balances)

    def _xp_mem(self, balances) -> List[int]:
        """
        Parameters
        ----------
        balances: List[int]
            The pool balances in native token units.

        Returns
        --------
        List[int]
            The pool balances in units of `D`.

        Note
        -----
        This intentionally always returns a new copy of the balances.
        """
        precisions = self.precisions
        price_scale = self.price_scale
        return [balances[0] * precisions[0]] + [
            balance * precision * price // PRECISION
            for balance, precision, price in zip(
                balances[1:], precisions[1:], price_scale
            )
        ]

    def _get_xcp(self, D: int) -> int:
        """
        Calculate the constant-product profit, using the balances at
        equilibrium point.
        """
        n_coins: int = self.n
        price_scale: List[int] = self.price_scale
        x = [D // n_coins] + [
            D * PRECISION // (price * n_coins) for price in price_scale
        ]
        return geometric_mean(x)

    def _increment_timestamp(self, blocks=1, timestamp=None) -> None:
        """Update the internal clock used to mimic the block timestamp."""
        if timestamp:
            self._block_timestamp = timestamp
            return

        self._block_timestamp += 12 * blocks

    # pylint: disable-next=R0912,R0913,R0914,R0915
    def _tweak_price(  # noqa: complexity: 12
        self,
        A: int,
        gamma: int,
        _xp: List[int],
        i: int,
        p_i: Optional[int],
        new_D: int,
        K0_prev: int = 0,
    ) -> None:
        """
        Applies several kinds of updates:
            - EMA price update: price_oracle
            - Profit counters: D, virtual_price, xcp_profit
            - price adjustment: price_scale
                - If p_i is None, the spot price will be used as the last price

        Also claims admin fees if appropriate (enough profit and price scale
        and oracle is close enough).
        """
        price_oracle: List[int] = self._price_oracle
        last_prices: List[int] = self.last_prices
        price_scale: List[int] = self.price_scale
        last_prices_timestamp: int = self.last_prices_timestamp
        block_timestamp: int = self._block_timestamp
        n_coins: int = self.n

        # Update EMA price oracle for a new block.  Happens once per block.
        # EMA uses price of the last trade and oracle price in previous block.
        if last_prices_timestamp < block_timestamp:
            ma_half_time: int = self.ma_half_time
            alpha: int = get_alpha(
                ma_half_time, block_timestamp, last_prices_timestamp, n_coins
            )
            if n_coins == 2:
                price_oracle = [
                    (last_p * (10**18 - alpha) + oracle_p * alpha) // 10**18
                    for last_p, oracle_p in zip(last_prices, price_oracle)
                ]
            elif n_coins == 3:
                # Cap state price that goes into the EMA with 2*price_scale.
                price_oracle = [
                    (min(last_p, 2 * p) * (10**18 - alpha) + oracle_p * alpha)
                    // 10**18
                    for last_p, oracle_p, p in zip(
                        last_prices, price_oracle, price_scale
                    )
                ]
            else:
                raise CalculationError("More than 3 coins is not supported.")
            self._price_oracle = price_oracle
            self.last_prices_timestamp = block_timestamp

        D_unadjusted: int = new_D  # Withdrawal methods know new D already
        if new_D == 0:
            D_unadjusted = newton_D(A, gamma, _xp, K0_prev)

        if p_i is None:
            if n_coins == 2:
                # calculate real prices
                __xp: List[int] = _xp.copy()
                dx_price: int = __xp[0] // 10**6
                __xp[0] += dx_price
                last_prices = [
                    price_scale[k - 1]
                    * dx_price
                    // (
                        __xp[k]
                        - factory_2_coin.newton_y(A, gamma, __xp, D_unadjusted, k)
                    )
                    for k in range(1, n_coins)
                ]
            else:
                last_prices = get_p(_xp, D_unadjusted, A, gamma)
                last_prices = [
                    last_p * p // 10**18
                    for last_p, p in zip(last_prices, price_scale)
                ]
        elif p_i > 0:
            # Save the last price
            if i > 0:
                last_prices[i - 1] = p_i
            else:
                # If 0th price changed - change all prices instead
                for k in range(n_coins - 1):
                    last_prices[k] = last_prices[k] * 10**18 // p_i
        else:
            raise CalculationError(f"p_i (last price) cannot be {p_i}.")

        self.last_prices = last_prices

        total_supply: int = self.tokens
        old_xcp_profit: int = self.xcp_profit
        old_virtual_price: int = self.virtual_price

        # Update profit numbers without price adjustment first
        xp: List[int] = [D_unadjusted // n_coins] + [
            D_unadjusted * PRECISION // (n_coins * price) for price in price_scale
        ]
        xcp_profit: int = 10**18
        virtual_price: int = 10**18

        if old_virtual_price > 0:
            xcp: int = geometric_mean(xp)
            virtual_price = 10**18 * xcp // total_supply

            if virtual_price < old_virtual_price:
                raise CryptoPoolError("Loss")

            xcp_profit = old_xcp_profit * virtual_price // old_virtual_price

        self.xcp_profit = xcp_profit

        if virtual_price * 2 - 10**18 > xcp_profit + 2 * self.allowed_extra_profit:
            norm: int = 0
            ratio: int = 0
            for k in range(n_coins - 1):
                ratio = price_oracle[k] * 10**18 // price_scale[k]
                ratio = abs(ratio - 10**18)
                norm += ratio**2
            norm = isqrt(norm)
            adjustment_step: int = max(self.adjustment_step, norm // 5)

            if norm > adjustment_step:
                new_prices = [
                    (p * (norm - adjustment_step) + adjustment_step * p_oracle) // norm
                    for p, p_oracle in zip(price_scale, price_oracle)
                ]

                # Calculate balances * prices
                xp = [_xp[0]] + [
                    balance * p_new // p
                    for balance, p, p_new in zip(_xp[1:], price_scale, new_prices)
                ]

                # Calculate "extended constant product" invariant xCP and virtual price
                D: int = newton_D(A, gamma, xp)
                xp = [D // n_coins] + [
                    D * PRECISION // (n_coins * p_new) for p_new in new_prices
                ]
                new_virtual_price = 10**18 * geometric_mean(xp) // total_supply

                # Proceed if we've got enough profit:
                #   new_virtual_price > 10**18
                #   new_virtual_price - 10**18 > (xcp_profit - 10**18) / 2
                if (new_virtual_price > 10**18) and (
                    2 * new_virtual_price - 10**18 > xcp_profit
                ):
                    self.price_scale = new_prices
                    self.D = D
                    self.virtual_price = new_virtual_price
                    return

        # If we are here, the price_scale adjustment did not happen
        # Still need to update the profit counter and D
        self.D = D_unadjusted
        self.virtual_price = virtual_price

    def _claim_admin_fees(self) -> None:
        """
        If the pool's profit has increased since the last fee claim, update profit,
        pool value, and LP token supply to reflect the admin taking its share of the
        fees by minting itself LP tokens. Otherwise, change nothing.

        Tricrypto-NG and Cryptopool implement this functionality differently, so we
        copy only Tricrypto-NG's way in this class for consistency.
        """
        # no gulping logic needed for the python code
        A: int = self.A
        gamma: int = self.gamma

        xcp_profit: int = self.xcp_profit
        xcp_profit_a: int = self.xcp_profit_a
        total_supply: int = self.tokens

        if xcp_profit <= xcp_profit_a or total_supply < 10**18:
            return

        vprice: int = self.virtual_price

        fees: int = (xcp_profit - xcp_profit_a) * self.admin_fee // (2 * 10**10)

        if fees > 0:
            frac: int = vprice * 10**18 // (vprice - fees) - 10**18
            d_supply: int = total_supply * frac // 10**18
            self.tokens += d_supply
            xcp_profit -= fees * 2
            self.xcp_profit = xcp_profit

        D: int = newton_D(A, gamma, self._xp())
        self.D = D

        self.virtual_price = 10**18 * self._get_xcp(D) // self.tokens
        self.xcp_profit_a = xcp_profit

    def get_dy(self, i: int, j: int, dx: int) -> int:
        """
        Calculate the amount received from swapping `dx`
        amount of the `i`-th coin for the `j`-th coin.

        Parameters
        ----------
        i: int
            Index of 'in' coin
        j: int
            Index of 'out' coin
        dx: int
            Amount of 'in' coin
        Returns
        -------
        int
            The 'out' coin amount

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        assert i != j  # dev: same input and output coin
        assert i < self.n  # dev: coin index out of range
        assert j < self.n  # dev: coin index out of range

        xp: List[int] = self.balances.copy()
        xp[i] += dx
        xp = self._xp_mem(xp)

        A: int = self.A
        gamma: int = self.gamma
        D: int = self.D

        y: int = get_y(A, gamma, xp, D, j)[0]
        dy: int = xp[j] - y - 1
        xp[j] = y
        precisions: List[int] = self.precisions
        price_scale: List[int] = self.price_scale
        if j > 0:
            dy = dy * PRECISION // (price_scale[j - 1] * precisions[j])
        else:
            dy = dy // precisions[0]
        dy -= self._fee(xp) * dy // 10**10

        return dy

    def get_y(self, i: int, j: int, x: int, xp: List[int]) -> int:
        r"""
        Calculate x[j] if one makes x[i] = x.

        Parameters
        ----------
        i: int
            index of coin; usually the "in"-token
        j: int
            index of coin; usually the "out"-token
        x: int
            balance of i-th coin in units of `D`
        xp: list of int
            coin balances in units of `D`

        Returns
        -------
        int
            The balance of the j-th coin, in units of `D`, for the other
            coin balances given.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A: int = self.A
        gamma: int = self.gamma
        D: int = newton_D(A, gamma, xp)

        xp = xp.copy()
        xp[i] = x

        y, _ = get_y(A, gamma, xp, D, j)
        return y

    def _fee(self, xp: List[int]) -> int:
        """
        f = fee_gamma / (fee_gamma + (1 - K))
        where
        K = prod(x) / (sum(x) / N)**N
        (all normalized to 1e18)
        """
        n_coins: int = self.n
        fee_gamma: int = self.fee_gamma
        if n_coins == 2:
            f: int = xp[0] + xp[1]
            f = (
                fee_gamma
                * 10**18
                // (
                    fee_gamma
                    + 10**18
                    - (10**18 * n_coins**n_coins) * xp[0] // f * xp[1] // f
                )
            )
        else:
            _sum_xp: int = sum(xp)
            K: int = 10**18
            for _x in xp:
                K = K * n_coins * _x // _sum_xp
            f = fee_gamma * 10**18 // (fee_gamma + 10**18 - K)
        return (self.mid_fee * f + self.out_fee * (10**18 - f)) // 10**18

    # pylint: disable-next=too-many-locals
    def _exchange(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int,
    ) -> Tuple[int, int]:
        assert i != j, "Indices must be different"
        assert i < self.n, "Index out of bounds"
        assert j < self.n, "Index out of bounds"
        assert dx > 0, "Can't swap zero amount"

        A = self.A
        gamma = self.gamma
        xp: List[int] = self.balances.copy()
        ix: int = j

        y: int = xp[j]
        xp[i] += dx
        self.balances[i] = xp[i]

        xp = self._xp_mem(xp)

        y_out = get_y(A, gamma, xp, self.D, j)
        dy: int = xp[j] - y_out[0]
        assert dy >= 0, f"Invalid dy: dx: {dx}, dy: {dy}, i: {i}, j: {j} "
        xp[j] -= dy
        dy -= 1

        price_scale: int = self.price_scale[j - 1]
        prec_i: int = self.precisions[i]
        prec_j: int = self.precisions[j]

        if j > 0:
            dy = dy * PRECISION // (price_scale)
        dy = dy // prec_j

        fee = self._fee(xp) * dy // 10**10
        dy -= fee
        assert dy >= min_dy, f"Slippage: dy: {dy}"
        y -= dy

        self.balances[j] = y

        y *= prec_j
        if j > 0:
            y = y * price_scale // PRECISION
        xp[j] = y

        p: Optional[int] = None
        K0_prev: int = 0
        if self.n == 2:
            if dx > 10**5 and dy > 10**5:
                _dx: int = dx * prec_i
                _dy: int = dy * prec_j
                if i != 0 and j != 0:
                    p = self.last_prices[i - 1] * _dx // _dy
                elif i == 0:
                    p = _dx * 10**18 // _dy
                else:  # j == 0
                    p = _dy * 10**18 // _dx
                    ix = i
        else:
            K0_prev = y_out[1]

        self._tweak_price(A, gamma, xp, ix, p, 0, K0_prev)

        return dy, fee

    def exchange(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int = 0,
    ) -> Tuple[int, int]:
        """
        Swap `dx` amount of the `i`-th coin for the `j`-th coin.

        Parameters
        ----------
        i: int
            'In' coin index
        j: int
            'Out' coin index
        dx: int
            'In' coin amount
        min_dy: int, optional
            Minimum 'out' coin amount required (default = 0)

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)

        Note
        -----
        In the vyper contract, there is an option to exchange using WETH or ETH.
        """
        return self._exchange(i, j, dx, min_dy)

    def exchange_underlying(
        self,
        i: int,
        j: int,
        dx: int,
        min_dy: int = 0,
    ) -> Tuple[int, int]:
        """
        In the vyper contract, this exchanges using ETH instead of WETH.
        In Curvesim, this is the same as `exchange`.
        """
        return self.exchange(i, j, dx, min_dy)

    # pylint: disable-next=too-many-locals, too-many-statements
    def add_liquidity(
        self,
        amounts: List[int],
        min_mint_amount: int = 0,
    ) -> int:
        """
        Add liquidity into the pool by depositing coins for LP tokens.

        Parameters
        ----------
        amounts: List[int]
            Deposit amounts.  At least one coin amount must be nonzero.
        min_mint_amount: int
            Minimum amount of LP tokens required (default = 0)

        Returns
        -------
        int
            Amount of LP tokens minted.
        """
        assert all(x >= 0 for x in amounts)
        assert sum(amounts) > 0

        A = self.A
        gamma = self.gamma
        n_coins: int = self.n

        xp_old: List[int] = self._xp_mem(self.balances)

        for i in range(n_coins):
            self.balances[i] += amounts[i]

        xp: List[int] = self._xp_mem(self.balances)
        amountsp: List[int] = [xp[i] - xp_old[i] for i in range(n_coins)]

        old_D: int = self.D
        D: int = newton_D(A, gamma, xp)

        d_token: int = 0
        token_supply: int = self.tokens
        if old_D > 0:
            d_token = token_supply * D // old_D - token_supply
        else:
            # sets initial virtual price to 1
            d_token = self._get_xcp(D)
        assert d_token > 0  # dev: nothing minted

        d_token_fee: int = 0
        if old_D > 0:
            d_token_fee = self._calc_token_fee(amountsp, xp) * d_token // 10**10 + 1
            d_token -= d_token_fee
            token_supply += d_token
            self.tokens += d_token

            p: Optional[int] = None
            ix: int = -1
            if n_coins == 2 and d_token > 10**5:
                nonzero_indices = [i for i, a in enumerate(amounts) if a != 0]
                if len(nonzero_indices) == 1:
                    # Calculate price for 2 coins:
                    # p_i * (dx_i - dtoken / token_supply * xx_i)
                    # = sum{k!=i}(p_k * (dtoken / token_supply * xx_k - dx_k))
                    # only ix is nonzero
                    prec: List[int] = self.precisions
                    last_prices: List[int] = self.last_prices
                    balances: List[int] = self.balances

                    ix = amounts.index(0)
                    S: int = 0
                    for i in range(n_coins):
                        if i == ix:
                            continue

                        if i == 0:
                            S += balances[i] * prec[i]
                        else:
                            S += balances[i] * prec[i] * last_prices[i - 1] // PRECISION
                    S = S * d_token // token_supply
                    p = (
                        S
                        * PRECISION
                        // (
                            amounts[ix] * prec[ix]
                            - d_token * balances[ix] * prec[ix] // token_supply
                        )
                    )

            self._tweak_price(A, gamma, xp, ix, p, D)

        else:
            self.D = D
            self.virtual_price = 10**18
            self.xcp_profit = 10**18
            self.xcp_profit_a = 10**18
            self.tokens += d_token

        assert d_token >= min_mint_amount, "Slippage"

        self._claim_admin_fees()

        return d_token

    def _calc_token_fee(self, amounts: List[int], xp: List[int]) -> int:
        n_coins: int = self.n
        # fee = sum(amounts_i - avg(amounts)) * fee' / sum(amounts)
        fee: int = self._fee(xp) * n_coins // (4 * (n_coins - 1))
        S: int = sum(amounts)
        avg: int = S // n_coins
        Sdiff: int = sum(abs(_x - avg) for _x in amounts)
        return fee * Sdiff // S + NOISE_FEE

    def remove_liquidity(
        self,
        _amount: int,
        min_amounts=None,
    ) -> List[int]:
        """
        Remove liquidity (burn LP tokens) to receive back part (or all) of
        the deposited funds.

        Parameters
        ----------
        _amount: int
            Amount LP tokens to burn.
        min_amounts: List[int], optional
            Minimum required amounts for each coin.  Default is 0 each.

        Returns
        -------
        List[int]
            The amounts of each coin received.

        Note
        ----
        "This withdrawal method is very safe, does no complex math"
        """
        n_coins: int = self.n
        min_amounts = min_amounts or [0] * n_coins

        amount: int = _amount
        balances: List[int] = self.balances
        withdraw_amounts: List[int] = [0] * n_coins

        self._claim_admin_fees()

        total_supply: int = self.tokens
        assert amount <= total_supply
        self.tokens -= amount

        if amount != total_supply:
            amount -= 1  # Make rounding errors favor other LPs a tiny bit

        for i in range(n_coins):
            withdraw_amounts[i] = balances[i] * amount // total_supply
            assert withdraw_amounts[i] >= min_amounts[i]
            self.balances[i] = balances[i] - withdraw_amounts[i]

        D: int = self.D
        self.D = D - D * amount // total_supply

        return withdraw_amounts

    def remove_liquidity_one_coin(
        self, token_amount: int, i: int, min_amount: int
    ) -> int:
        """
        Remove liquidity entirely in one type of coin.
        Fees will be extracted and there may be significant price impact incurred.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to burn.
        i: int
            Index of the `out` coin.
        min_amount: int
            Minimum amount of the 'out' coin required (default = 0)

        Returns
        -------
        int
            Amount of the `i`-th coin received.
        """
        A: int = self.A
        gamma: int = self.gamma

        dy: int = 0
        D: int = 0
        p: Optional[int] = None
        xp: List[int] = [0] * self.n

        self._claim_admin_fees()

        calc_price: bool = self.n == 2

        dy, p, D, xp = self._calc_withdraw_one_coin(
            A, gamma, token_amount, i, False, calc_price
        )
        assert dy >= min_amount, "Slippage"

        self.balances[i] -= dy
        self.tokens -= token_amount

        self._tweak_price(A, gamma, xp, i, p, D)

        return dy

    # pylint: disable-next=too-many-locals,too-many-arguments, too-many-branches
    def _calc_withdraw_one_coin(
        self,
        A: int,
        gamma: int,
        token_amount: int,
        i: int,
        update_D: bool,
        calc_price: bool,
    ) -> Tuple[int, Optional[int], int, List[int]]:
        """
        Calculate the output amount from burning `token amount` of LP tokens
        and receiving entirely in the `i`-th coin.

        Parameters
        ----------
        A: int
            Amplification coefficient equal to `A * n**n` from the whitepaper.
        gamma: int
            Gamma coefficient from the whitepaper.
        token_amount: int
            Amount of LP tokens to burn.
        i: int
            Index of the `out` coin.
        update_D: bool
            Switch for recomputing `D` during calculation.
        calc_price: bool
            Switch for recomputing the price of token `i` after simulating a
            withdrawal. Only does something if `n` = 2.

        Returns
        -------
        int
            Output amount of the `i`-th coin.
        Optional[int]
            Price of the `i`-th coin after the withdrawal.
        int
            `D` after the withdrawal, accounting for fee.
        List[int]
            `xp` after the withdrawal, accounting for fee.

        Note
        ----
        The 3-coin pool contract doesn't have calc_price even though the 2-coin one
        does, so the price value returned is `None` when the price calculation doesn't
        occur.

        This is a "view" function; it doesn't change the state of the pool.
        """
        token_supply: int = self.tokens
        assert token_amount <= token_supply
        assert i < self.n, "Index out of bounds"

        xx: List[int] = self.balances.copy()
        precisions: List[int] = self.precisions

        xp: List[int] = self._xp_mem(xx)

        if update_D:
            D0: int = newton_D(A, gamma, xp)
        else:
            D0 = self.D

        D: int = D0

        if self.n == 2:
            fee: int = self._fee(xp)
        elif self.n == 3:
            # For n = 3, charge max fee if xp_correction > xp_imprecise[i]
            # Specifically, if % of xp[i] withdrawn > 1/n = 1/3
            # Otherwise, _fee() underflows
            n_coins: int = self.n
            xp_imprecise: List[int] = xp.copy()
            xp_correction: int = xp[i] * n_coins * token_amount // token_supply

            if xp_correction < xp_imprecise[i]:
                xp_imprecise[i] -= xp_correction
                fee = self._fee(xp_imprecise)
            else:
                fee = self.out_fee
        else:
            raise CalculationError(
                "_calc_withdraw_one_coin doesn't support more than 3 coins"
            )

        # Charge fee on D, not on y, e.g. reducing invariant LESS than charging user
        dD: int = token_amount * D // token_supply
        D_fee: int = fee * dD // (2 * 10**10) + 1
        D -= dD - D_fee
        y: int = get_y(A, gamma, xp, D, i)[0]
        if i == 0:
            dy: int = (xp[i] - y) // precisions[i]
        else:
            dy = (xp[i] - y) * PRECISION // (precisions[i] * self.price_scale[i - 1])
        xp[i] = y

        # Price calc
        p: Optional[int] = None
        if self.n == 2 and calc_price:
            if dy > 10**5 and token_amount > 10**5:
                # p_i = dD / D0 * sum'(p_k * x_k) / (dy - dD / D0 * y0)
                j = (i + 1) % 2
                precision: int = precisions[i]
                S: int = xx[j] * precisions[j]
                S = S * dD // D0
                p = S * PRECISION // (dy * precision - dD * xx[i] * precision // D0)
                if i == 0:
                    p = (10**18) ** 2 // p

        return dy, p, D, xp

    def calc_withdraw_one_coin(self, token_amount: int, i: int) -> int:
        """
        Calculate the output amount from burning `token amount` of LP tokens
        and receiving entirely in the `i`-th coin.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to burn.
        i: int
            Index of the `out` coin.

        Returns
        -------
        int
            Output amount of the `i`-th coin.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        return self._calc_withdraw_one_coin(
            self.A, self.gamma, token_amount, i, True, False
        )[0]

    def lp_price(self) -> int:
        """
        Returns the price of an LP token in units of token 0.

        Derived from the equilibrium point of a constant-product AMM
        that approximates Cryptoswap's behavior.

        Returns
        -------
        int
            Liquidity redeemable per LP token in units of token 0.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        if self.n == 2:
            virtual_price: int = self.virtual_price
            price_oracle: List[int] = self.internal_price_oracle()
            price: int = factory_2_coin.lp_price(virtual_price, price_oracle)
        elif self.n == 3:
            # 3-coin vyper contract uses cached packed oracle prices
            # instead of internal_price_oracle()
            virtual_price = self.virtual_price
            price_oracle = self._price_oracle
            price = tricrypto_ng.lp_price(virtual_price, price_oracle)
        else:
            raise CalculationError("LP price calc doesn't support more than 3 coins")

        return price

    def internal_price_oracle(self) -> List[int]:
        """
        Return the value of the EMA price oracle.
        """
        price_oracle: List[int] = self._price_oracle
        last_prices_timestamp: int = self.last_prices_timestamp

        block_timestamp: int = self._block_timestamp
        if last_prices_timestamp < block_timestamp:

            if self.n == 2:
                last_prices: List[int] = self.last_prices
            elif self.n == 3:
                # 3-coin vyper contract caps every "last price" that gets fed to the EMA
                last_prices = self.last_prices.copy()
                price_scale: List[int] = self.price_scale
                last_prices = [
                    min(last_p, 2 * storage_p)
                    for last_p, storage_p in zip(last_prices, price_scale)
                ]
            else:
                raise CalculationError(
                    "Price oracle calc doesn't support more than 3 coins"
                )

            ma_half_time: int = self.ma_half_time
            n_coins: int = self.n
            alpha: int = get_alpha(
                ma_half_time, block_timestamp, last_prices_timestamp, n_coins
            )

            return [
                (last_p * (10**18 - alpha) + oracle_p * alpha) // 10**18
                for last_p, oracle_p in zip(last_prices, price_oracle)
            ]

        return price_oracle

    def price_oracle(self) -> List[int]:
        """
        Return the value of the EMA price oracle.

        Same as `internal_price_oracle`.  Kept for compatability with the
        vyper interface.
        """
        return self.internal_price_oracle()

    def get_virtual_price(self) -> int:
        """
        Return the virtual price of an LP token.
        """
        return 10**18 * self._get_xcp(self.D) // self.tokens

    def calc_token_amount(self, amounts: List[int]) -> int:
        """
        Calculate the amount of LP tokens minted by depositing given amounts.

        Parameters
        ----------
        amounts: List[int]
            Deposit amounts.  At least one coin amount must be nonzero.

        Returns
        -------
        int
            Amount of LP tokens minted.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        token_supply: int = self.tokens
        A: int = self.A
        gamma: int = self.gamma
        xp: List[int] = self._xp()
        amountsp: List[int] = self._xp_mem(amounts)
        D0: int = self.D

        for i, a in enumerate(amountsp):
            xp[i] += a

        D: int = newton_D(A, gamma, xp)
        d_token: int = token_supply * D // D0 - token_supply
        d_token -= self._calc_token_fee(amountsp, xp) * d_token // 10**10 + 1

        return d_token

    def dydxfee(self, i, j):
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Trading fees are deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin with fees deducted.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        return self.dydx(i, j, use_fee=True)

    def dydx(self, i, j, use_fee=False):  # pylint: disable=too-many-locals
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Defaults to no fees deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        xp = self._xp()
        x_i = xp[i]
        x_j = xp[j]
        n = len(xp)

        D = self.D
        A = self.A
        A_multiplier = 10**4
        gamma = self.gamma

        K0 = 10**18 * n**n * prod(xp) / D**n

        coeff = A * gamma**2 / (10**18 + gamma - K0) ** 2
        frac = (10**18 + gamma + K0) * (sum(xp) - D) / (10**18 + gamma - K0)
        dydx_top = x_j * (A_multiplier * D + coeff * (x_i + frac))
        dydx_bottom = x_i * (A_multiplier * D + coeff * (x_j + frac))
        dydx = dydx_top / dydx_bottom

        if j > 0:
            price_scale = self.price_scale[j - 1]
            dydx = dydx * 10**18 / price_scale
        if i > 0:
            price_scale = self.price_scale[i - 1]
            dydx = dydx * price_scale / 10**18

        if use_fee:
            fee = self._fee(xp)
            dydx = dydx - dydx * fee / 10**10

        return dydx


def _get_unix_timestamp():
    """Get the timestamp in Unix time."""
    return int(time.time())



"""
Implements the volume-limited arbitrage pipeline.
"""

import os

from curvesim.iterators.param_samplers import ParameterizedPoolIterator
from curvesim.iterators.price_samplers import PriceVolume
from curvesim.logging import get_logger
from curvesim.metrics import init_metrics, make_results
from curvesim.pool import get_sim_pool
from curvesim.pool_data import get_pool_volume

from .. import run_pipeline
from ..common import DEFAULT_METRICS
from .strategy import VolumeLimitedStrategy

logger = get_logger(__name__)


# pylint: disable-next=too-many-locals
def pipeline(
    pool_metadata,
    *,
    variable_params=None,
    fixed_params=None,
    metrics=None,
    days=60,
    src="coingecko",
    data_dir="data",
    vol_mult=None,
    ncpu=None,
    end=None,
):
    """
    Implements the volume-limited arbitrage pipeline.

    At each timestep, the pool is arbitraged as close to the prevailing market price
    as possible without surpassing a volume constraint. By default, volume is limited
    to the total market volume at each timestep, scaled by the proportion of
    volume attributable to the pool over the whole simulation period (vol_mult).

    Parameters
    ----------
    pool_metadata : :class:`~curvesim.pool_data.metadata.PoolMetaDataInterface`
        Pool metadata object for the pool of interest.

    variable_params : dict, defaults to broad range of A/fee values
        Pool parameters to vary across simulations.
        keys: pool parameters, values: iterables of ints

        Example
        --------
        >>> variable_params = {"A": [100, 1000], "fee": [10**6, 4*10**6]}

    fixed_params : dict, optional
        Pool parameters set before all simulations.
        keys: pool parameters, values: ints

        Example
        --------
        >>> fixed_params = {"D": 1000000*10**18}

    days : int, default=60
        Number of days to pull pool and price data for.

    src : str, default="coingecko"
        Source for price/volume data: "coingecko" or "local".

    data_dir : str, default="data"
        relative path to saved price data folder

    vol_mult : dict, default computed from data
        Value(s) multiplied by market volume to specify volume limits
        (overrides vol_mode).

        dict should map from trade-pair tuples to values, e.g.:

        .. code-block::

            {('DAI', 'USDC'): 0.1, ('DAI', 'USDT'): 0.1, ('USDC', 'USDT'): 0.1}

    ncpu : int, default=os.cpu_count()
        Number of cores to use.

    Returns
    -------
    SimResults object
    """
    if ncpu is None:
        cpu_count = os.cpu_count()
        ncpu = cpu_count if cpu_count is not None else 1

    pool = get_sim_pool(pool_metadata)

    # pylint: disable-next=abstract-class-instantiated
    param_sampler = ParameterizedPoolIterator(pool, variable_params, fixed_params)
    price_sampler = PriceVolume(
        pool.assets, days=days, data_dir=data_dir, src=src, end=end
    )

    if vol_mult is None:
        pool_volume = get_pool_volume(pool_metadata, days=days, end=end)
        vol_mult = pool_volume.sum() / price_sampler.volumes.sum()
        logger.info("Volume Multipliers:\n%s", vol_mult.to_string())
        vol_mult = vol_mult.to_dict()

    metrics = metrics or DEFAULT_METRICS
    metrics = init_metrics(metrics, pool=pool)
    strategy = VolumeLimitedStrategy(metrics, vol_mult)

    output = run_pipeline(param_sampler, price_sampler, strategy, ncpu=ncpu)
    results = make_results(*output, metrics)

    return results



"""
Utility functions for working with datetimes in UTC format.

This module provides custom versions of common datetime operations to ensure
returned datetimes are always in the UTC timezone and are not naive.

The main goal is to prevent potential issues related to timezone-aware and
timezone-naive datetime objects.

Recommended style for importing is:

    from curvesim.utils import datetime

    datetime.now()

This syntactically substitutes the `datetime.datetime` class methods with
this module's functions.
"""
from datetime import datetime, timezone


def now() -> datetime:
    """
    Customized version of `datetime.datetime.now` to ensure
    UTC timezone and format.

    Returns
    -------
    datetime
        Current UTC datetime with hour, minute, second, and microsecond set to 0.
        This datetime is timezone-aware and is not naive.
    """
    utc_dt = datetime.now(timezone.utc).replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    return utc_dt


def fromtimestamp(timestamp: int) -> datetime:
    """
    Customized version of `datetime.datetime.fromtimestamp` to ensure
    UTC timezone and format.

    Parameters
    ----------
    timestamp : int
        Unix timestamp to be converted to datetime.

    Returns
    -------
    datetime
        The UTC datetime representation of the given timestamp.
        This datetime is timezone-aware and is not naive.
    """
    utc_timestamp = datetime.fromtimestamp(timestamp, tz=timezone.utc)
    return utc_timestamp



"""
Mainly a module to house the `Pool`, a basic stableswap implementation in Python.
"""
from math import prod
from typing import Type

from gmpy2 import mpz

from curvesim.exceptions import CurvesimValueError
from curvesim.pool.snapshot import CurvePoolBalanceSnapshot, Snapshot

from ..base import Pool


class CurvePool(Pool):  # pylint: disable=too-many-instance-attributes
    """
    Basic stableswap implementation in Python.
    """

    snapshot_class: Type[Snapshot] = CurvePoolBalanceSnapshot

    __slots__ = (
        "A",
        "n",
        "fee",
        "rates",
        "balances",
        "tokens",
        "fee_mul",
        "admin_fee",
        "r",
        "n_total",
        "admin_balances",
    )

    def __init__(  # pylint: disable=too-many-arguments
        self,
        A,
        D,
        n,
        rates=None,
        tokens=None,
        fee=4 * 10**6,
        fee_mul=None,
        admin_fee=5 * 10**9,
        virtual_price=None,
    ):
        """
        Parameters
        ----------
        A : int
            Amplification coefficient; this is :math:`A n^{n-1}` in the whitepaper.
        D : int or list of int
            virtual total balance or pool coin balances in native token units
        n: int
            number of coins
        rates: list of int
            precision and rate adjustments
        tokens: int
            LP token supply
        fee: int, optional
            fee with 10**10 precision (default = .004%)
        fee_mul: optional
            fee multiplier for dynamic fee pools
        admin_fee: int, optional
            percentage of `fee` with 10**10 precision (default = 50%)
        virtual_price: int, optional
            amount of D invariant per LP token; can be used when
            missing `tokens` value.
        """
        rates = rates or [10**18] * n

        self.A = A
        self.n = n
        self.fee = fee
        self.rates = rates

        if isinstance(D, list):
            self.balances = D.copy()
        else:
            self.balances = self._convert_D_to_balances(D)

        if tokens and virtual_price:
            raise CurvesimValueError(
                "Should not set both `tokens` and `virtual_price`."
            )

        # By now, should have set everything needed for D.
        D = self.D()
        if tokens:
            self.tokens = tokens
        elif virtual_price:
            self.tokens = D * 10**18 // virtual_price
        else:
            self.tokens = D

        self.fee_mul = fee_mul
        self.admin_fee = admin_fee
        self.r = False
        self.n_total = n
        self.admin_balances = [0] * n

    def _convert_D_to_balances(self, D):
        n = self.n
        rates = self.rates
        return [D // n * 10**18 // _p for _p in rates]

    def _xp(self):
        rates = self.rates
        balances = self.balances
        return self._xp_mem(rates, balances)

    def _xp_mem(self, rates, balances):
        return [x * p // 10**18 for x, p in zip(balances, rates)]

    def D(self, xp=None):
        """
        `D` is the stableswap invariant; this can be thought of as the value of
        all coin balances if the pool were to become balanced.

        Convenience wrapper for `get_D` which uses the set `A` and makes `xp`
        an optional arg.

        Parameters
        ----------
        xp: list of ints
            Coin balances in units of D

        Returns
        -------
        int
            The stableswap invariant, `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A = self.A
        xp = xp or self._xp()
        return self.get_D(xp, A)

    def get_D(self, xp, A):
        r"""
        Calculate D invariant iteratively using non-overflowing integer operations.

        Stableswap equation:

        .. math::
             A n^n \sum{x_i} + D = A n^n D + D^{n+1} / (n^n \prod{x_i})

        Converging solution using Newton's method:

        .. math::
             d_{j+1} = (A n^n \sum{x_i} + n d_j^{n+1} / (n^n \prod{x_i}))
                     / (A n^n + (n+1) d_j^n/(n^n \prod{x_i}) - 1)

        Replace :math:`A n^n` by `An` and :math:`d_j^{n+1}/(n^n \prod{x_i})` by
        :math:`D_p` to arrive at the iterative formula in the code.

        Parameters
        ----------
        xp: list of ints
            Coin balances in units of D
        A: int
            Amplification coefficient

        Returns
        -------
        int
            The stableswap invariant, `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """  # noqa
        Dprev = 0
        S = sum(xp)
        D = S
        n = self.n
        Ann = A * n
        D = mpz(D)
        Ann = mpz(Ann)
        while abs(D - Dprev) > 1:
            D_P = D
            for x in xp:
                D_P = D_P * D // (n * x)
            Dprev = D
            D = (Ann * S + D_P * n) * D // ((Ann - 1) * D + (n + 1) * D_P)

        D = int(D)
        return D

    def get_D_mem(self, balances, A):
        """
        Convenience wrapper for `get_D` which takes in balances in token units.
        Naming is based on the vyper equivalent.

        Parameters
        ----------
        balances: list of ints
            Coin balances in native token units
        A: int
            Amplification coefficient

        Returns
        -------
        int
            The stableswap invariant, `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        xp = [x * p // 10**18 for x, p in zip(balances, self.rates)]
        return self.get_D(xp, A)

    def get_y(self, i, j, x, xp):
        r"""
        Calculate x[j] if one makes x[i] = x.

        The stableswap equation gives the following:

        .. math::
            x_1^2 + x_1 (\operatorname{sum'} - (A n^n - 1) D / (A n^n))
               = D^{n+1}/(n^{2 n} \operatorname{prod'} A)

        where :math:`\operatorname{sum'}` is the sum of all :math:`x_i` for
        :math:`i \\neq j` and :math:`\operatorname{prod'}` is the product
        of all :math:`x_i` for :math:`i \\neq j`.

        This is a quadratic equation in :math:`x_j`.

        .. math:: x_1^2 + b x_1 = c

        which can then be solved iteratively by Newton's method:

        .. math:: x_1 := (x_1^2 + c) / (2 x_1 + b)

        Parameters
        ----------
        i: int
            index of coin; usually the "in"-token
        j: int
            index of coin; usually the "out"-token
        x: int
            balance of i-th coin in units of D
        xp: list of int
            coin balances in units of D

        Returns
        -------
        int
            The balance of the j-th coin, in units of D, for the other
            coin balances given.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """  # noqa
        xx = xp[:]
        D = self.D(xx)
        D = mpz(D)
        xx[i] = x  # x is quantity of underlying asset brought to 1e18 precision
        n = self.n
        xx = [xx[k] for k in range(n) if k != j]
        Ann = self.A * n
        c = D
        for y in xx:
            c = c * D // (y * n)
        c = c * D // (n * Ann)
        b = sum(xx) + D // Ann - D
        y_prev = 0
        y = D
        while abs(y - y_prev) > 1:
            y_prev = y
            y = (y**2 + c) // (2 * y + b)
        y = int(y)
        return y  # result is in units for D

    def get_y_D(self, A, i, xp, D):
        """
        Calculate x[i] if one uses a reduced `D` than one calculated for given `xp`.

        See docstring for `get_y`.

        Parameters
        ----------
        A: int
            Amplification coefficient for given xp and D
        i: int
            index of coin to calculate balance for
        xp: list of int
            coin balances in units of D
        D: int
            new invariant value

        Returns
        -------
        int
            The balance of the i-th coin, in units of D

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        D = mpz(D)
        n = self.n
        xx = [xp[k] for k in range(n) if k != i]
        S = sum(xx)
        Ann = A * n
        c = D
        for y in xx:
            c = c * D // (y * n)
        c = c * D // (n * Ann)
        b = S + D // Ann
        y_prev = 0
        y = D
        while abs(y - y_prev) > 1:
            y_prev = y
            y = (y**2 + c) // (2 * y + b - D)
        y = int(y)
        return y  # result is in units for D

    def exchange(self, i, j, dx):
        """
        Perform an exchange between two coins.
        Index values can be found via the `coins` public getter method.

        Parameters
        ----------
        i : int
            Index of "in" coin.
        j : int
            Index of "out" coin.
        dx : int
            Amount of coin `i` being exchanged.

        Returns
        -------
        (int, int)
            (amount of coin `j` received, trading fee)


        Examples
        --------

        >>> pool = Pool(A=250, D=1000000 * 10**18, n=2, p=[10**30, 10**30])
        >>> pool.exchange(0, 1, 150 * 10**6)
        (149939820, 59999)
        """
        xp = self._xp()
        x = xp[i] + dx * self.rates[i] // 10**18
        y = self.get_y(i, j, x, xp)
        dy = xp[j] - y - 1

        if self.fee_mul is None:
            fee = dy * self.fee // 10**10
        else:
            fee = dy * self.dynamic_fee((xp[i] + x) // 2, (xp[j] + y) // 2) // 10**10

        admin_fee = fee * self.admin_fee // 10**10

        # Convert all to real units
        rate = self.rates[j]
        dy = (dy - fee) * 10**18 // rate
        fee = fee * 10**18 // rate
        admin_fee = admin_fee * 10**18 // rate
        assert dy >= 0

        self.balances[i] += dx
        self.balances[j] -= dy + admin_fee
        self.admin_balances[j] += admin_fee
        return dy, fee

    # pylint: disable-next=too-many-locals
    def calc_withdraw_one_coin(self, token_amount, i, use_fee=True):
        """
        Calculate the amount in the i-th coin received from
        redeeming the given amount of LP tokens.

        By default, fees are deducted.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to redeem
        i: int
            Index of coin to withdraw in.
        use_fee: bool, default=True
            Deduct fees.

        Returns
        -------
        int
            Redemption amount in i-th coin

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A = self.A
        xp = self._xp()
        D0 = self.D()
        D1 = D0 - token_amount * D0 // self.tokens

        new_y = self.get_y_D(A, i, xp, D1)
        dy_before_fee = (xp[i] - new_y) * 10**18 // self.rates[i]

        xp_reduced = xp
        if self.fee and use_fee:
            n_coins = self.n
            _fee = self.fee * n_coins // (4 * (n_coins - 1))

            for j in range(n_coins):
                dx_expected = 0
                if j == i:
                    dx_expected = xp[j] * D1 // D0 - new_y
                else:
                    dx_expected = xp[j] - xp[j] * D1 // D0
                xp_reduced[j] -= _fee * dx_expected // 10**10

        dy = xp[i] - self.get_y_D(A, i, xp_reduced, D1)
        dy = (dy - 1) * 10**18 // self.rates[i]

        if use_fee:
            dy_fee = dy_before_fee - dy
            return dy, dy_fee

        return dy

    def add_liquidity(self, amounts):
        """
        Deposit coin amounts for LP token.

        Parameters
        ----------
        amounts: list of int
            Coin amounts to deposit

        Returns
        -------
        int
            LP token amount received for the deposit amounts.
        """
        mint_amount, fees = self.calc_token_amount(amounts, use_fee=True)
        self.tokens += mint_amount

        balances = self.balances
        afee = self.admin_fee
        admin_fees = [f * afee // 10**10 for f in fees]
        new_balances = [
            bal + amt - fee for bal, amt, fee in zip(balances, amounts, admin_fees)
        ]
        self.balances = new_balances
        self.admin_balances = [t + a for t, a in zip(self.admin_balances, admin_fees)]

        return mint_amount

    def remove_liquidity_one_coin(self, token_amount, i):
        """
        Redeem given LP token amount for the i-th coin.

        Parameters
        ----------
        token_amount: int
            Amount of LP tokens to redeem
        i: int
            Index of coin to withdraw in

        Returns
        -------
        int
            Redemption amount in i-th coin
        """
        dy, dy_fee = self.calc_withdraw_one_coin(token_amount, i, use_fee=True)
        admin_fee = dy_fee * self.admin_fee // 10**10
        self.balances[i] -= dy + admin_fee
        self.admin_balances[i] += admin_fee
        self.tokens -= token_amount
        return dy, dy_fee

    def remove_liquidity(self, _amount, min_amounts=None):
        """
        Remove liquidity (burn LP tokens) to receive back part (or all) of
        the deposited funds.

        Parameters
        ----------
        _amount: int
            Amount LP tokens to burn.
        min_amounts: List[int], optional
            Minimum required amounts for each coin.  Default is 0 each.

        Note
        ----
        "This withdrawal method is very safe, does no complex math"
        """
        min_amounts = min_amounts or [0] * self.n

        total_supply = self.tokens
        self.tokens -= _amount
        balances = self.balances

        for i in range(self.n):
            d_balance = balances[i] * _amount // total_supply
            assert d_balance >= min_amounts[i]
            self.balances[i] = balances[i] - d_balance

    # pylint: disable-next=too-many-locals
    def remove_liquidity_imbalance(self, amounts, max_burn_amount=None):
        """
        Withdraw an imbalanced amount of tokens from the pool. Accounts for fees.

        Parameters
        ----------
        amounts: List[int]
            Amounts of tokens to withdraw (positive ints)
        max_burn_amount: int, optional
            Maximum amount of LP tokens to burn

        Returns
        -------
        burn_amount: int
            Amount of LP token burned in the withdrawal
        fees: List[int]
            Amount of fees paid
        """
        A = self.A
        old_balances = self.balances
        D0 = self.get_D_mem(old_balances, A)

        new_balances = self.balances[:]
        for i in range(self.n):
            amount = amounts[i]
            assert amount >= 0  # must input positive ints
            new_balances[i] -= amount
        D1 = self.get_D_mem(new_balances, A)

        fees = [0] * self.n
        _fee = self.fee * self.n // (4 * (self.n - 1))
        for i in range(self.n):
            ideal_balance = D1 * old_balances[i] // D0
            difference = abs(ideal_balance - new_balances[i])
            fees[i] = _fee * difference // 10**10
            admin_fee = fees[i] * self.admin_fee // 10**10
            self.admin_balances[i] += admin_fee
            self.balances[i] = new_balances[i] - admin_fee  # sans admin fees
            new_balances[i] -= fees[i]

        D2 = self.get_D_mem(new_balances, A)

        burn_amount = self.tokens * (D0 - D2) // D0 + 1  # should be positive
        if max_burn_amount:
            assert burn_amount <= max_burn_amount
        assert burn_amount >= 0

        self.tokens -= burn_amount

        return burn_amount, fees

    # pylint: disable-next=too-many-locals
    def calc_token_amount(self, amounts, use_fee=False):
        """
        Calculate the amount of LP tokens received for the given coin
        deposit amounts.

        Fee logic is based on add_liquidity, which makes this more accurate than
        the `calc_token_amount` in the actual contract, which neglects fees.

        By default, it's assumed you the same behavior as the vyper contract,
        which is to NOT deduct fees.

        Parameters
        ----------
        amounts: list of int
            Coin amounts to be deposited.
        use_fee: bool, default=False
            Deduct fees.

        Returns
        -------
        int
            LP token amount received for the deposit amounts.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        A = self.A
        old_balances = self.balances
        D0 = self.get_D_mem(old_balances, A)

        new_balances = self.balances[:]
        for i in range(self.n):
            new_balances[i] += amounts[i]
        D1 = self.get_D_mem(new_balances, A)

        mint_balances = new_balances[:]

        if use_fee:
            _fee = self.fee * self.n // (4 * (self.n - 1))

            fees = [0] * self.n
            for i in range(self.n):
                ideal_balance = D1 * old_balances[i] // D0
                difference = abs(ideal_balance - new_balances[i])
                fees[i] = _fee * difference // 10**10
                mint_balances[i] -= fees[i]

        D2 = self.get_D_mem(mint_balances, A)

        mint_amount = self.tokens * (D2 - D0) // D0

        if use_fee:
            return mint_amount, fees

        return mint_amount

    def get_virtual_price(self):
        """
        Returns the expected value of one LP token if the pool were
        to become perfectly balanced (all coins revert to peg).

        Returns
        -------
        int
            Amount of the stableswap invariant, `D`, corresponding to
            one LP token, in units of `D`.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        return self.D() * 10**18 // self.tokens

    def dynamic_fee(self, xpi, xpj):
        xps2 = xpi + xpj
        xps2 *= xps2  # Doing just ** 2 can overflow apparently
        return (self.fee_mul * self.fee) // (
            (self.fee_mul - 10**10) * 4 * xpi * xpj // xps2 + 10**10
        )

    def dydxfee(self, i, j):
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Trading fees are deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin with fees deducted.

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        return self.dydx(i, j, use_fee=True)

    def dydx(self, i, j, use_fee=False):
        """
        Returns the spot price of i-th coin quoted in terms of j-th coin,
        i.e. the ratio of output coin amount to input coin amount for
        an "infinitesimally" small trade.

        Defaults to no fees deducted.

        Parameters
        ----------
        i: int
            Index of coin to be priced; in a swapping context, this is
            the "in"-token.
        j: int
            Index of quote currency; in a swapping context, this is the
            "out"-token.

        Returns
        -------
        float
            Price of i-th coin quoted in j-th coin

        Note
        ----
        This is a "view" function; it doesn't change the state of the pool.
        """
        xp = self._xp()
        return self._dydx(i, j, xp, use_fee)

    def _dydx(self, i, j, xp, use_fee):
        xi = xp[i]
        xj = xp[j]
        n = self.n
        A = self.A
        D = self.D(xp)
        D_pow = mpz(D) ** (n + 1)
        x_prod = prod(xp)
        A_pow = A * n ** (n + 1)
        dydx = (xj * (xi * A_pow * x_prod + D_pow)) / (
            xi * (xj * A_pow * x_prod + D_pow)
        )

        if use_fee:
            if self.fee_mul is None:
                fee_factor = self.fee / 10**10
            else:
                fee_factor = self.dynamic_fee(xi, xj) / 10**10
        else:
            fee_factor = 0

        dydx *= 1 - fee_factor

        return float(dydx)



"""Utlity functions for general usage in Curvesim."""
__all__ = [
    "get_env_var",
    "get_pairs",
    "dataclass",
    "get_event_loop",
    "cache",
    "override",
    "datetime",
    "is_address",
    "to_address",
    "Address",
]

import asyncio
import os
import sys
from dataclasses import dataclass as _dataclass
from itertools import combinations

from dotenv import load_dotenv

from curvesim.exceptions import MissingEnvVarError

from .address import Address, is_address, to_address
from .decorators import cache, override

load_dotenv()

is_py310 = sys.version_info.minor >= 10 or sys.version_info.major > 3
"""
Some utils require knowing the python version.  Right now it's only important to
distinguish between 3.10 and earlier versions.
"""

_NOT_VALUE = object()
"""
Dummy value for optional default arg
so that any value, including `None`,
can be set as a default.
"""


def get_env_var(var_name, default=_NOT_VALUE):
    """
    Retrieve environment variable.

    Parameters
    ----------
    var_name: str
        Name of the environment variable.
    default: object
        Value to return if env var is missing.

    Returns
    -------
    str
        Value of the environment variable.

    Raise
    -----
    curvesim.exception.MissingEnvVarError
        Raised if default is not set and env var is missing.
    """
    var_value = os.getenv(var_name)
    if var_value is None:
        if default is _NOT_VALUE:
            raise MissingEnvVarError(f"Could not get env var: '{var_name}'")
        return default

    return var_value


def get_pairs(arg):
    """
    Get sorted pairwise combinations of an iterable.
    Integer inputs are treated as range(int).

    Parameters
    ----------
    arg: iterable or int
        The iterable to produce pairs from, or an integer specifying the range
        to produce pairs from.

    Returns
    -------
    list
        Sorted pairwise combinations of the input.
    """
    if isinstance(arg, int):
        arg = range(arg)

    return list(combinations(arg, 2))


def dataclass(*args, **kwargs):
    """
    Slightly modified version of the standard library's `dataclass`.

    The modification is to allow the setting of slots on versions of
    python before 3.10.

    Right now, we just remove the `slots` kwarg if it exists, but in the
    future we can implement our own custom slots for old versions.
    """
    if "slots" in kwargs and not is_py310:
        del kwargs["slots"]

    return _dataclass(*args, **kwargs)


def get_event_loop():
    """
    Access the event loop without using asyncio.get_event_loop().

    Generally, you should run scheduled coroutines soon after calling
    this to avoid overwriting event loops that have unrun coroutines.

    Calling asyncio.get_event_loop() when an event loop isn't running and/or
    set will cause a DeprecationWarning in various versions of Python 3.10-3.12,
    and a future release will start raising an error instead:
    https://docs.python.org/3.11/library/asyncio-eventloop.html?highlight=selectoreventloop#asyncio.get_event_loop.

    Implementation slightly modified from https://stackoverflow.com/a/73884759
    as below works for all versions >= 3.7.
    """
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    return loop



"""
Tools for fetching pool metadata and volume.

Currently supports stableswap pools, metapools, rebasing (RAI) metapools,
and 2-token cryptopools.
"""

__all__ = ["get_metadata", "get_pool_volume"]


from .queries.metadata import get_metadata
from .queries.pool_volume import get_pool_volume



from curvesim.pool_data import get_pool_volume
from curvesim.pool_data.metadata import PoolMetaData
from curvesim.utils import get_pairs

from ..unit.test_pool_metadata import (
    cryptopool_test_metadata,
    metapool_test_metadata,
    pool_test_metadata,
    tricrypto_ng_test_metadata,
)


def test_get_pool_volume():
    """Test the volume query."""
    metadata_list = [
        cryptopool_test_metadata,
        metapool_test_metadata,
        pool_test_metadata,
        tricrypto_ng_test_metadata,
    ]

    for metadata in metadata_list:
        pool_metadata = PoolMetaData(metadata)

        # Test using metadata
        volumes1 = get_pool_volume(pool_metadata, days=2, end=1698292800)
        assert len(volumes1) == 2
        assert volumes1.columns.to_list() == get_pairs(pool_metadata.coin_names)

        # Test using address and chain
        address = pool_metadata.address
        chain = pool_metadata.chain
        volumes2 = get_pool_volume(address, chain=chain, days=2, end=1698292800)
        assert len(volumes2) == 2
        assert volumes2.columns.to_list() == get_pairs(pool_metadata.coin_names)

        assert all(volumes1 == volumes2)



"""Unit tests for the SnapshotMixin and derived subclasses of Snapshot."""
import pytest

from curvesim.exceptions import SnapshotError
from curvesim.pool.sim_interface import SimCurveMetaPool, SimCurvePool
from curvesim.pool.snapshot import SnapshotMixin


def test_snapshot_raises_exception():
    """Test error thrown for missing snapshot class."""
    obj = SnapshotMixin()
    with pytest.raises(SnapshotError):
        obj.get_snapshot()


def test_pool_balance_snapshot():
    """Test balances are copied and reset properly for SimCurvePool."""
    pool = SimCurvePool(A=250, D=1000000 * 10**18, n=2, admin_fee=5 * 10**9)

    snapshot = pool.get_snapshot()

    pre_balances = pool.balances.copy()
    pre_admin_balances = pool.admin_balances.copy()
    pool.exchange(0, 1, 10**12)
    post_balances = pool.balances.copy()
    post_admin_balances = pool.admin_balances.copy()
    assert pre_balances != post_balances
    assert pre_admin_balances != post_admin_balances

    pool.revert_to_snapshot(snapshot)
    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances


def test_metapool_balance_snapshot():
    """Test balances are copied and reset properly for SimCurveMetaPool."""
    basepool = SimCurvePool(A=1000, D=2750000 * 10**18, n=2, admin_fee=5 * 10**9)
    pool = SimCurveMetaPool(
        A=250, D=4000000 * 10**18, n=2, admin_fee=5 * 10**9, basepool=basepool
    )

    snapshot = pool.get_snapshot()

    pre_balances = pool.balances.copy()
    pre_bp_balances = basepool.balances.copy()
    pre_admin_balances = pool.admin_balances.copy()
    pre_bp_admin_balances = basepool.admin_balances.copy()
    pre_lp_tokens = basepool.tokens

    # ---- test top-level exchange ---- #

    pool.exchange(0, 1, 10**12)

    # `exchange` changes values for metapool not basepool
    assert pool.balances != pre_balances
    assert pool.admin_balances != pre_admin_balances
    assert basepool.balances == pre_bp_balances
    assert basepool.admin_balances == pre_bp_admin_balances
    assert basepool.tokens == pre_lp_tokens

    pool.revert_to_snapshot(snapshot)

    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances
    assert basepool.balances == pre_bp_balances
    assert basepool.admin_balances == pre_bp_admin_balances

    # -- test exchange between primary and basepool underlyer -- #

    pool.exchange_underlying(0, 1, 157 * 10**18)

    # `exchange_underlying` should change basepool values
    assert pool.balances != pre_balances
    assert pool.admin_balances != pre_admin_balances
    assert basepool.balances != pre_bp_balances
    assert basepool.admin_balances != pre_bp_admin_balances
    assert basepool.tokens != pre_lp_tokens

    pool.revert_to_snapshot(snapshot)

    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances
    assert basepool.balances == pre_bp_balances
    assert basepool.admin_balances == pre_bp_admin_balances
    assert basepool.tokens == pre_lp_tokens


def test_2_coin_cryptopool_balance_snapshot(sim_curve_crypto_pool):
    _test_cryptool_balance_snapshot(sim_curve_crypto_pool)


def test_tricrypto_balance_snapshot(sim_curve_tricrypto_pool):
    _test_cryptool_balance_snapshot(sim_curve_tricrypto_pool)


def _test_cryptool_balance_snapshot(pool):
    snapshot = pool.get_snapshot()

    pre_balances = pool.balances.copy()
    pre_D = pool.D
    pre_price_scale = pool.price_scale.copy()
    pre_price_oracle = pool._price_oracle.copy()
    pre_virtual_price = pool.virtual_price
    pre_xcp_profit = pool.xcp_profit
    pre_xcp_profit_a = pool.xcp_profit_a
    pre_last_prices = pool.last_prices.copy()
    pre_last_prices_timestamp = pool.last_prices_timestamp

    # ---- test exchange ---- #
    # Use two trades since last_prices == _price_oracle at start
    pool._increment_timestamp(42)  # 504 seconds
    pool.exchange(0, 1, 10**26 // 2)
    pool._increment_timestamp(42)
    pool.exchange(0, 1, 10**26 // 2)

    # test that `exchange` changes expected values
    assert pool.balances != pre_balances
    assert pool.D != pre_D
    assert pool.price_scale != pre_price_scale  # _tweak_price should update price_scale
    assert pool._price_oracle != pre_price_oracle
    assert pool.virtual_price != pre_virtual_price
    assert pool.xcp_profit != pre_xcp_profit
    assert pool.xcp_profit_a == pre_xcp_profit_a  # not claimed yet
    assert pool.last_prices != pre_last_prices
    assert pool.last_prices_timestamp != pre_last_prices_timestamp

    # test _claim_admin_fees
    pool._claim_admin_fees()
    assert pool.xcp_profit_a != pre_xcp_profit_a

    # Ensure all reverted
    pool.revert_to_snapshot(snapshot)
    assert pool.balances == pre_balances
    assert pool.D == pre_D
    assert pool.price_scale == pre_price_scale
    assert pool._price_oracle == pre_price_oracle
    assert pool.virtual_price == pre_virtual_price
    assert pool.xcp_profit == pre_xcp_profit
    assert pool.xcp_profit_a == pre_xcp_profit_a
    assert pool.last_prices == pre_last_prices
    assert pool.last_prices_timestamp == pre_last_prices_timestamp


def test_snapshot_context():
    """Test context manager."""
    pool = SimCurvePool(A=250, D=1000000 * 10**18, n=2, admin_fee=5 * 10**9)

    pre_balances = pool.balances.copy()
    pre_admin_balances = pool.admin_balances.copy()

    with pool.use_snapshot_context():
        pool.exchange(0, 1, 10**12)
        assert pool.balances != pre_balances
        assert pool.admin_balances != pre_admin_balances

    assert pool.balances == pre_balances
    assert pool.admin_balances == pre_admin_balances



"""
Pool subpackage for creating Curve pool objects.

This subpackage can be used standalone from the simulation functionality
for interactive exploration of Curve pool behavior. The two primary ways to create
pools are `get` and `make`, which allow you to create pools from on-chain state or p
assed-in parameters, respectively.

Pool calculations are thoroughly unit-tested against the corresponding
smart contract code.  We aim for the exact results as in the vyper
integer arithmetic. The pool interfaces largely adhere to the smart contracts
but in a few cases allow an extra option, such as enabling/disabling fees.
"""

__all__ = [
    "get",
    "make",
    "Pool",
    "CurvePool",
    "CurveMetaPool",
    "CurveRaiPool",
    "CurveCryptoPool",
    "SimCurvePool",
    "SimCurveMetaPool",
    "SimCurveRaiPool",
]

from curvesim.exceptions import CurvesimValueError
from curvesim.logging import get_logger
from curvesim.pool_data import get_metadata
from curvesim.pool_data.metadata import PoolMetaData, PoolMetaDataInterface

from .base import Pool
from .cryptoswap import CurveCryptoPool
from .sim_interface import SimCurveMetaPool, SimCurvePool, SimCurveRaiPool
from .stableswap import CurveMetaPool, CurvePool, CurveRaiPool

logger = get_logger(__name__)


def make(
    A,
    D,
    n,
    *,
    basepool=None,
    rates=None,
    rate_multiplier=None,
    tokens=None,
    fee=4 * 10**6,
    fee_mul=None,
    admin_fee=0 * 10**9,
):
    """
    Factory function for creating pools from "raw" parameters, i.e. no
    data pulled from chain.

    Parameters
    ----------
    A: int
        Amplification coefficient.  This controls the curvature of the
        stableswap bonding curve.  Increased values makes the curve
        flatter in a greater neighborhood of equal balances.

        Defaults to [int(2 ** (a / 2)) for a in range(12, 28)].

    D: int
        Total pool liquidity given in 18 decimal precision.

    n: int
        The number of token-types in the pool (e.g., DAI, USDC, USDT = 3)

    basepool: dict, optional
        a dict cointaining the arguments for instantiating a basepool

    rates: list of int, optional
        precisions for each coin

    rate_multiplier: int, optional
        precision and rate adjustment for primary stable in metapool

    tokens: int, optional
        Total LP token supply.

        .. note::
            The number of tokens does not influence "normal" pool computations;
            but, for metapools, the number of basepool tokens is critically
            important to trade calculations.

    fee: int
        Fees taken for both liquidity providers and the DAO.

        Units are in fixed-point so that 10**10 is 100%,
        e.g. 4 * 10**6 is 4 bps and 2 * 10**8 is 2%.

    fee_mul : int
        fee multiplier for dynamic fee pools

    admin_fee : int, default=0 * 10**9
        Fees taken for the DAO.  For factory pools, it is half of the total fees,
        as was typical for previous non-factory pools.

        Units are fixed-point percentage of `fee`, e.g. 5 * 10**9 is
        50% of the total fees.

    Returns
    -------
    :class:`Pool`
    """
    if rates and rate_multiplier:
        raise CurvesimValueError("Should have only `rates` or `rate_multiplier`.")

    if basepool:
        pool = CurveMetaPool(
            A,
            D,
            n,
            basepool,
            rate_multiplier=rate_multiplier,
            tokens=tokens,
            fee=fee,
            fee_mul=fee_mul,
            admin_fee=admin_fee,
        )

    else:
        pool = CurvePool(
            A,
            D,
            n,
            rates=rates,
            tokens=tokens,
            fee=fee,
            fee_mul=fee_mul,
            admin_fee=admin_fee,
        )

    return pool


def get_pool(
    pool_metadata,
    chain="mainnet",
    *,
    normalize=False,
    end_ts=None,
    env="prod",
):
    """
    Factory function for creating a pool based on metadata pulled from on-chain.

    Parameters
    ----------
    pool_metadata : Union[str, dict, PoolMetaDataInterface]
        pool address prefixed with "0x" or already pulled metadata in the form
        of a dict or :class:`PoolMetaDataInterface`.

    chain: str, default="mainnet"
        chain/layer2 identifier, e.g. "mainnet", "arbitrum", "optimism"

    normalize : bool, default=False
        If True, normalizes balances to 18 decimals (useful for sim calculations).

    end_ts: int, optional
        Posix timestamp indicating the datetime of the metadata snapshot.
        Only used when `pool_metadata` is an address.

    Returns
    -------
    :class:`Pool`

    Examples
    --------
    >>> import curvesim
    >>> pool_address = "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"
    >>> chain = "mainnet"
    >>> pool = curvesim.pool.get(pool_address, chain)
    """
    if end_ts and not isinstance(pool_metadata, str):
        raise CurvesimValueError("`end_ts` has no effect unless pool address is used.")

    if isinstance(pool_metadata, str):
        pool_metadata = get_metadata(pool_metadata, chain=chain, env=env, end_ts=end_ts)
    elif isinstance(pool_metadata, dict):
        pool_metadata = PoolMetaData(pool_metadata)

    if not isinstance(pool_metadata, PoolMetaDataInterface):
        raise CurvesimValueError(
            "`pool_metadata` must be of type `str`, `dict`, or `PoolMetaDataInterface`."
        )

    init_kwargs = pool_metadata.init_kwargs(normalize)
    logger.debug(init_kwargs)

    pool_type = pool_metadata.pool_type
    pool = pool_type(**init_kwargs)

    pool.metadata = pool_metadata._dict  # pylint: disable=protected-access

    return pool


POOL_TYPE_TO_CUSTOM_KWARGS = {SimCurveRaiPool: ["redemption_prices"]}


def get_sim_pool(
    pool_metadata,
    chain="mainnet",
    *,
    balanced=True,
    balanced_base=True,
    custom_kwargs=None,
    end_ts=None,
    env="prod",
):
    """
    Factory function for creating a sim pool based on metadata pulled from on-chain.

    Parameters
    ----------
    pool_metadata : Union[str, dict, PoolMetaDataInterface]
        pool address prefixed with "0x" or already pulled metadata in the form
        of a dict or :class:`PoolMetaDataInterface`.

    chain: str, default="mainnet"
        chain/layer2 identifier, e.g. "mainnet", "arbitrum", "optimism"

    balanced : bool, default=True
        If True, balances the pool value across assets.

    balanced_base : bool, default=True
        If True and pool is metapool, balances the basepool value across assets.

    end_ts: int, optional
        Posix timestamp indicating the datetime of the metadata snapshot.
        Only used when `pool_metadata` is an address.

    custom_kwargs: dict, optional
        Used for passing additional kwargs to the pool's `__init__`.

    Returns
    -------
    :class:`SimPool`

    Note
    -----
    The balances are always normalized to 18 decimals.

    Examples
    --------
    >>> import curvesim
    >>> pool_address = "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"
    >>> chain = "mainnet"
    >>> pool = curvesim.pool.get_sim_pool(pool_address, chain)
    """
    custom_kwargs = custom_kwargs or {}

    if end_ts and not isinstance(pool_metadata, str):
        raise CurvesimValueError("`end_ts` has no effect unless pool address is used.")

    if isinstance(pool_metadata, str):
        pool_metadata = get_metadata(pool_metadata, chain=chain, env=env, end_ts=end_ts)
    elif isinstance(pool_metadata, dict):
        pool_metadata = PoolMetaData(pool_metadata)

    if not isinstance(pool_metadata, PoolMetaDataInterface):
        raise CurvesimValueError(
            "`pool_metadata` must be of type `str`, `dict`, or `PoolMetaDataInterface`."
        )

    init_kwargs = pool_metadata.init_kwargs(normalize=True)
    logger.debug(init_kwargs)

    pool_type = pool_metadata.sim_pool_type

    custom_keys = POOL_TYPE_TO_CUSTOM_KWARGS.get(pool_type, [])
    for key in custom_keys:
        try:
            init_kwargs[key] = custom_kwargs.get(key)
        except KeyError as e:
            raise CurvesimValueError(f"'{pool_type.__name__}' needs '{key}'.") from e

    pool = pool_type(**init_kwargs)
    pool.metadata = pool_metadata._dict  # pylint: disable=protected-access
    _balance_pool(pool, balanced, balanced_base)

    return pool


def _balance_pool(pool, balanced, balanced_base):
    """
    Balances the pool and/or its basepool if applicable.

    Note: Mutates the `pool` argument.
    """
    # pylint: disable=protected-access
    if balanced:
        if callable(pool.D):
            # stableswap/metapool
            D = pool.D()
        else:
            # cryptopool
            D = pool.D
        pool.balances = pool._convert_D_to_balances(D)

    if balanced_base and hasattr(pool, "basepool"):
        basepool = pool.basepool
        D = basepool.D()
        basepool.balances = basepool._convert_D_to_balances(D)


get = get_pool



"""
Constants and Enum types used in Curvesim.
"""
from enum import Enum


class StrEnum(str, Enum):
    """
    Custom string enum type since the builtin `StrEnum` is not available
    until Python 3.11.
    """

    def __str__(self):
        """
        Regular Enum's __str__ is the name, rather than the value,
        e.g.

        >>> str(Chain.MAINNET)
        'Chain.MAINNET'

        so we need to explicitly use the value.

        This behaves like the builtin `StrEnum` (available in 3.11).
        """
        return str.__str__(self)


class Chain(StrEnum):
    """Identifiers for chains & layer 2s."""

    MAINNET = "mainnet"
    ARBITRUM = "arbitrum"
    OPTIMISM = "optimism"
    FANTOM = "fantom"
    AVALANCHE = "avalanche"
    MATIC = "matic"
    XDAI = "xdai"


class Env(StrEnum):
    """Names for different API environments."""

    PROD = "prod"
    STAGING = "staging"



from abc import ABC, abstractmethod


class Log(ABC):
    @abstractmethod
    def update(self, **kwargs):
        """Updates log data with event data."""

    @abstractmethod
    def compute_metrics(self):
        """Computes metrics from the accumulated log data."""



from math import comb

from curvesim.pool_data.queries.pool_volume import _get_pair_data


class DummyMetadata:
    def __init__(self, address, coin_names, coin_addresses, n, basepool_address=None):
        self.address = address
        self.coin_names = coin_names
        self.coins = coin_addresses
        self.n = n
        self._dict = {
            "basepool": {"address": basepool_address},
            "coins": coin_addresses,
        }


def make_dummy_metadata(n):
    if isinstance(n, list):
        n_total = sum(n) - 1
    else:
        n_total = n

    kwargs = {
        "address": "pool_address",
        "coin_names": ["SYM" + str(i) for i in range(n_total)],
        "coin_addresses": ["ADR" + str(i) for i in range(n_total)],
        "n": n,
        "basepool_address": "basepool_address",
    }

    return DummyMetadata(**kwargs), n_total


def test_get_pair_data():
    for n in range(2, 5):
        pool_metadata, n_total = make_dummy_metadata(n)
        pair_data = _get_pair_data(pool_metadata)
        pair_data = tuple(pair_data)
        pool_addresses, pair_addresses, pair_symbols = zip(*pair_data)

        # Ensure correct length
        assert len(pair_data) == comb(n_total, 2)

        # Ensure correct pool address
        assert pool_addresses == ("pool_address",) * comb(n_total, 2)

        # Ensure all pairs unique
        assert len(pair_addresses) == len(set(pair_addresses))  # all pairs unique
        assert len(pair_symbols) == len(set(pair_symbols))  # all pairs unique

        # Ensure all items present
        assert {x for pair in pair_addresses for x in pair} == set(pool_metadata.coins)
        assert {x for pair in pair_symbols for x in pair} == set(
            pool_metadata.coin_names
        )


def test_get_pair_data_metapool():
    n_list = [[n1, n2] for n1 in range(2, 4) for n2 in range(2, 5)]
    for n in n_list:
        pool_metadata, n_total = make_dummy_metadata(n)
        pair_data = _get_pair_data(pool_metadata)
        pool_addresses, pair_addresses, pair_symbols = zip(*pair_data)

        # Ensure correct length
        assert len(pair_data) == comb(n_total, 2)

        # Ensure correct pool addresses
        n_meta = n[0] - 1
        n_base = n[1]
        n_pairs_meta = comb(n_meta, 2) + n_meta * n_base
        n_pairs_base = comb(n_base, 2)
        assert (
            pool_addresses
            == ("pool_address",) * n_pairs_meta + ("basepool_address",) * n_pairs_base
        )

        # Ensure all pairs unique
        assert len(pair_addresses) == len(set(pair_addresses))  # all pairs unique
        assert len(pair_symbols) == len(set(pair_symbols))  # all pairs unique

        # Ensure all items present
        assert {x for pair in pair_addresses for x in pair} == set(pool_metadata.coins)
        assert {x for pair in pair_symbols for x in pair} == set(
            pool_metadata.coin_names
        )



from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st

from curvesim.pipelines.vol_limited_arb.trader import _apply_volume_limits
from curvesim.templates.trader import ArbTrade


class DummyPool:
    def __init__(self, min_trade_sizes):
        self.min_trade_sizes = min_trade_sizes

    def get_min_trade_size(self, coin_in):
        return self.min_trade_sizes[coin_in]


random_int = st.integers(min_value=1, max_value=100)
three_ints = st.tuples(random_int, random_int, random_int)


@given(three_ints, three_ints, three_ints)
@settings(
    max_examples=10,
    deadline=None,
    suppress_health_check=[HealthCheck.function_scoped_fixture],
)
def test_apply_volume_limits(trade_sizes, min_trade_sizes, volume_limits):
    coins = ["SYM0", "SYM1", "SYM2"]
    pairs = [("SYM0", "SYM1"), ("SYM0", "SYM2"), ("SYM1", "SYM2")]

    arb_trades = [ArbTrade(*pair, size, 0) for pair, size in zip(pairs, trade_sizes)]
    limits = dict(zip(pairs, volume_limits))
    pool = DummyPool(dict(zip(coins, min_trade_sizes)))

    limited_arb_trades, excluded_trades = _apply_volume_limits(arb_trades, limits, pool)

    for trade in limited_arb_trades:
        assert trade.amount_in <= limits[trade.coin_in, trade.coin_out]
        assert trade.amount_in > pool.get_min_trade_size(trade.coin_in)

    for trade in excluded_trades:
        assert trade.amount_in <= pool.get_min_trade_size(trade.coin_in)



from curvesim.version import parse_version


def test_parse_version():
    """
    Check the version string gets correctly parsed to 'info' tuple.
    """
    version_tuple = parse_version("0.3.5")
    expected_result = (0, 3, 5, "release", 0)
    assert version_tuple == expected_result

    version_tuple = parse_version("0.3.5.release")
    expected_result = (0, 3, 5, "release", 0)
    assert version_tuple == expected_result

    version_tuple = parse_version("0.3.5.rc3")
    expected_result = (0, 3, 5, "rc", 3)
    assert version_tuple == expected_result

    version_tuple = parse_version("0.3.5.rc0")
    expected_result = (0, 3, 5, "rc", 0)
    assert version_tuple == expected_result

    version_tuple = parse_version("0.3.5.a12")
    expected_result = (0, 3, 5, "a", 12)
    assert version_tuple == expected_result



from curvesim.logging import get_logger

from .base import PoolMetaDataBase

logger = get_logger(__name__)


class CryptoswapMetaData(PoolMetaDataBase):
    """Specific implementation of the `PoolMetaDataInterface` for Cryptoswap."""

    def init_kwargs(self, normalize=True):
        data = self._dict

        kwargs = {
            "A": data["params"]["A"],
            "gamma": data["params"]["gamma"],
            "n": len(data["coins"]["names"]),
            "mid_fee": data["params"]["mid_fee"],
            "out_fee": data["params"]["out_fee"],
            "allowed_extra_profit": data["params"]["allowed_extra_profit"],
            "fee_gamma": data["params"]["fee_gamma"],
            "adjustment_step": data["params"]["adjustment_step"],
            "price_scale": data["params"]["price_scale"],
            "admin_fee": data["params"]["admin_fee"],
            "ma_half_time": data["params"]["ma_half_time"],
            "virtual_price": data["reserves"]["virtual_price"],
            "xcp_profit": data["params"]["xcp_profit"],
            "xcp_profit_a": data["params"]["xcp_profit_a"],
        }
        n = kwargs["n"]

        if not normalize:
            kwargs["precisions"] = [10 ** (18 - d) for d in data["coins"]["decimals"]]
        else:
            kwargs["precisions"] = [1] * n

        if normalize:
            coin_balances = data["reserves"]["by_coin"]
        else:
            coin_balances = data["reserves"]["unnormalized_by_coin"]
        kwargs["balances"] = coin_balances

        return kwargs

    @property
    def coins(self):
        return self._dict["coins"]["addresses"]

    @property
    def coin_names(self):
        return self._dict["coins"]["names"]

    @property
    def n(self):
        return len(self._dict["coins"]["names"])



"""
Functions to get historical volume for Curve pools.
"""

from datetime import datetime, timezone
from math import comb
from typing import List, Optional, Tuple, Union

from pandas import DataFrame, Series

from curvesim.logging import get_logger
from curvesim.network.curve_prices import get_pool_pair_volume_sync
from curvesim.pool_data.metadata import PoolMetaDataInterface
from curvesim.utils import get_event_loop, get_pairs

from .metadata import Chain, get_metadata

logger = get_logger(__name__)


def get_pool_volume(
    metadata_or_address: Union[PoolMetaDataInterface, str],
    days: int = 60,
    end: Optional[int] = None,
    chain: Union[str, Chain] = "mainnet",
) -> DataFrame:
    """
    Gets historical daily volume for each pair of coins traded in a Curve pool.

    Parameters
    ----------
    metadata_or_address: PoolMetaDataInterface or str
        Pool metadata or pool address to fetch metadata.

    days: int, defaults to 60
        Number of days to pull volume data for.

    end: int, defaults to start of current date
        Posix timestamp of the last time to pull data for.

    chain: str, default "mainnet"
        Chain to use if pool address is provided to fetch metadata.

    Returns
    -------
    DataFrame
        Rows: DateTimeIndex, Columns: pairwise tuples of coin symbols

    """

    logger.info("Fetching historical pool volume...")

    if isinstance(metadata_or_address, str):
        pool_metadata: PoolMetaDataInterface = get_metadata(metadata_or_address, chain)
    else:
        pool_metadata = metadata_or_address

    pair_data = _get_pair_data(pool_metadata)
    start_ts, end_ts = _process_timestamps(days, end)
    loop = get_event_loop()

    volumes: dict[Tuple[str, str], Series] = {}
    for pool_address, pair_addresses, pair_symbols in pair_data:
        data: DataFrame = get_pool_pair_volume_sync(
            pool_address,
            *pair_addresses,
            start_ts,
            end_ts,
            chain=pool_metadata.chain,
            event_loop=loop,
        )
        volumes[pair_symbols] = data["volume"]

    volume_df = _make_volume_df(volumes, days)
    return volume_df


def _get_pair_data(pool_metadata) -> List[Tuple[str, Tuple[str, str], Tuple[str, str]]]:
    coin_addresses = _get_coin_addresses(pool_metadata)
    pair_symbols = get_pairs(pool_metadata.coin_names)
    pair_addresses = get_pairs(coin_addresses)

    if isinstance(pool_metadata.n, list):
        pool_addresses = _get_metapool_addresses(pool_metadata)
    else:
        pool_addresses = [pool_metadata.address] * comb(pool_metadata.n, 2)
    return list(zip(pool_addresses, pair_addresses, pair_symbols))


def _get_coin_addresses(pool_metadata) -> List[str]:
    # pylint: disable=protected-access
    if "wrapper" in pool_metadata._dict["coins"]:
        return pool_metadata._dict["coins"]["wrapper"]["addresses"]

    return pool_metadata.coins


def _get_metapool_addresses(pool_metadata) -> List[str]:
    n_meta = pool_metadata.n[0] - 1
    n_base = pool_metadata.n[1]
    address_meta = pool_metadata.address
    # pylint: disable-next=protected-access
    address_base = pool_metadata._dict["basepool"]["address"]

    n_pairs_meta = comb(n_meta, 2) + n_meta * n_base
    n_pairs_base = comb(n_base, 2)

    return [address_meta] * n_pairs_meta + [address_base] * n_pairs_base


def _process_timestamps(days, end) -> Tuple[int, int]:
    end = end or int(
        datetime.now(timezone.utc)
        .replace(hour=0, minute=0, second=0, microsecond=0)
        .timestamp()
    )
    start = end - days * 86400
    return start, end


def _make_volume_df(volumes, days) -> DataFrame:
    df = DataFrame(volumes)
    df.columns = df.columns.to_flat_index()
    if len(df) > days:
        df = df[-days:]
    logger.info("Days of volume returned:\n%s", df.count().to_string())
    df.fillna(0, inplace=True)
    return df



import functools
import inspect
import re

from curvesim.exceptions import CurvesimException


def cache(user_function, /):
    """
    Simple lightweight unbounded cache.  Sometimes called "memoize".

    Returns the same as lru_cache(maxsize=None), creating a thin wrapper
    around a dictionary lookup for the function arguments. Because it
    never needs to evict old values, this is smaller and faster than
    lru_cache() with a size limit.

    The cache is threadsafe so the wrapped function can be used in
    multiple threads.

    ----
    This isn't in functools until python 3.9, so we copy over the
    implementation as in:
    https://github.com/python/cpython/blob/3.11/Lib/functools.py#L648
    """
    return functools.lru_cache(maxsize=None)(user_function)


def override(method):
    """
    Method decorator to signify and check a method overrides a method
    in a super class.

    Implementation taken from https://stackoverflow.com/a/14631397/1175053
    """
    stack = inspect.stack()
    base_classes = re.search(r"class.+\((.+)\)\s*\:", stack[2][4][0]).group(1)

    # handle multiple inheritance
    base_classes = [s.strip() for s in base_classes.split(",")]
    if not base_classes:
        raise CurvesimException("override decorator: unable to determine base class")

    # stack[0]=overrides, stack[1]=inside class def'n, stack[2]=outside class def'n
    derived_class_locals = stack[2][0].f_locals

    # replace each class name in base_classes with the actual class type
    for i, base_class in enumerate(base_classes):

        if "." not in base_class:
            base_classes[i] = derived_class_locals[base_class]

        else:
            components = base_class.split(".")

            # obj is either a module or a class
            obj = derived_class_locals[components[0]]

            for c in components[1:]:
                assert inspect.ismodule(obj) or inspect.isclass(obj)
                obj = getattr(obj, c)

            base_classes[i] = obj

    if not any(hasattr(cls, method.__name__) for cls in base_classes):
        raise CurvesimException(
            f'Overridden method "{method.__name__}" was not found in any super class.'
        )
    return method



from numpy import inf, mean

from curvesim.metrics.metrics import _compute_liquidity_density
from curvesim.utils import get_pairs


def test_liquidity_density_sim_curve_pool(sim_curve_pool):
    _test_liquidity_density_stableswap(sim_curve_pool)


def test_liquidity_density_sim_curve_tripool(sim_curve_tripool):
    _test_liquidity_density_stableswap(sim_curve_tripool)


def test_liquidity_density_sim_curve_meta_pool(sim_curve_meta_pool):
    basepool = sim_curve_meta_pool.basepool
    coin_names = ["BP_SYM" + str(i) for i in range(basepool.n)]
    basepool.metadata = {"coins": {"names": coin_names}}
    _test_liquidity_density_stableswap(sim_curve_meta_pool)


def test_liquidity_density_sim_curve_crypto_pool(sim_curve_crypto_pool):
    _test_liquidity_density_cryptoswap(sim_curve_crypto_pool)


def test_liquidity_density_sim_curve_tricrypto_pool(sim_curve_tricrypto_pool):
    _test_liquidity_density_cryptoswap(sim_curve_tricrypto_pool)


def _test_liquidity_density_stableswap(pool):
    """
    Tests liquidity density for stableswap pools:
    - Ensure LD is max at center
    - Ensure LDs ~equal at equivalent curve positions for every trading pair/direction
    - Ensure LD at center is ~correct
    """

    def trade_size_function(pool, coin_in):
        x_per_dx = 10**8
        return pool.asset_balances[coin_in] // x_per_dx

    A_list = [10, 50, 100, 500, 1000, 5000]
    for A in A_list:
        pool.A = A
        LD_range = _test_compute_liquidity_density(pool, trade_size_function)

        # Ensure LD at center is ~correct
        LD_max_expected = (A + 1) / 2
        assert abs(LD_range[0] - LD_max_expected) / LD_max_expected < 0.001


def _test_liquidity_density_cryptoswap(pool):
    """
    Tests liquidity density for cryptoswap pools:
    - Ensure LD is max at center
    - Ensure LDs ~equal at equivalent curve positions for every trading pair/direction
    - Ensure LD at center is ~correct
    - Ensure LD near tail is ~.5 (constant product LD)
    """

    def trade_size_function(pool, coin_in):
        return pool.get_min_trade_size(coin_in)

    pool.allowed_extra_profit = inf  # disable _tweak_price
    pool.balances = pool._convert_D_to_balances(pool.D)  # balance pool
    n = pool.n

    A_list = [10, 50, 100, 500]
    gamma_list = [10**x for x in range(13, 16)]
    for A in A_list:
        for gamma in gamma_list:
            pool.A = A * n**n * 10000
            pool.gamma = gamma
            LD_range = _test_compute_liquidity_density(pool, trade_size_function)

            # Ensure LD at center is ~correct
            A_v1 = A * n ** (n - 1)
            LD_max_expected = (A_v1 + 1) / 2
            assert abs(LD_range[0] - LD_max_expected) / LD_max_expected < 0.0015

            # Ensure LD near tail is ~.5 (constant product LD)
            assert 0.47 < LD_range[-1] < 0.501


def _test_compute_liquidity_density(pool, trade_size_fn):
    """
    Computes liquidity density ranges for each trading pair/direction and tests that:
    - LD is max at center
    - LDs ~equal at equivalent curve positions
    """
    coin_names = ["SYM" + str(i) for i in range(pool.n)]
    pool.metadata = {"coins": {"names": coin_names}}

    # Compute LD range for each trading pair/direction
    coin_pairs = get_pairs(pool.coin_names)
    LD_ranges = []
    for pair in coin_pairs:
        coin_in, coin_out = pair
        LD0 = _compute_liquidity_density_range(pool, coin_in, coin_out, trade_size_fn)
        LD1 = _compute_liquidity_density_range(pool, coin_out, coin_in, trade_size_fn)
        LD_ranges += [LD0, LD1]

    # Ensure LD is max at center
    for LD_range in LD_ranges:
        assert LD_range[0] == max(LD_range)

    # Ensure LDs ~equal at equivalent curve positions for every trading pair/direction
    LD_ranges_aligned = list(zip(*LD_ranges))
    LD_ranges_means = [mean(LDs) for LDs in LD_ranges_aligned]
    for LDs, LD_mean in zip(LD_ranges_aligned, LD_ranges_means):
        percent_deviations = [abs(LD - LD_mean) / LD_mean for LD in LDs]
        assert max(percent_deviations) < 1e-4

    return LD_ranges_means


def _compute_liquidity_density_range(pool, coin_in, coin_out, trade_size_function):
    """Computes liquidity density at a range of locations along the bonding curve."""

    pre_trade_size = 0
    pre_trade_step_size = pool.asset_balances[coin_in] // 250

    coin_out_balances = pool.asset_balances[coin_out]
    coin_out_balances_limit = coin_out_balances * 0.05

    LD_range = []
    while coin_out_balances > coin_out_balances_limit:
        with pool.use_snapshot_context():
            if pre_trade_size > 0:
                pool.trade(coin_in, coin_out, pre_trade_size)
            LD_trade_size = trade_size_function(pool, coin_in)
            LD = _compute_liquidity_density(pool, coin_in, coin_out, LD_trade_size)
            coin_out_balances = pool.asset_balances[coin_out]

        LD_range.append(LD)
        pre_trade_size += pre_trade_step_size

    return LD_range



"""
Contains the bonding_curve function, which computes a pool's bonding
curve and current reserves for each pair of coins and optionally
plots the curves using Matplotlib.
"""
from itertools import combinations
from typing import Dict, List, Tuple, Union

import matplotlib.pyplot as plt
from numpy import linspace

from curvesim.pool import CurveCryptoPool, CurveMetaPool, CurvePool, CurveRaiPool

D_UNIT = 10**18

Stableswap = Union[CurvePool, CurveMetaPool, CurveRaiPool]
Cryptoswap = Union[CurveCryptoPool]

IndexPair = Tuple[int, int]
NormalizedPoint = Tuple[int, int]
Point = Tuple[float, float]


# pylint: disable-next=too-many-locals
def bonding_curve(  # noqa: C901
    pool: Union[Stableswap, Cryptoswap], *, truncate=None, resolution=1000, plot=False
) -> Dict[IndexPair, List[Point]]:
    """
    Computes and optionally plots a pool's bonding curve and current reserves.

    Parameters
    ----------
    pool : CurvePool, CurveMetaPool, CurveRaiPool, or CurveCryptoPool
        The pool object for which the bonding curve is computed.

    truncate : float, int, or None, optional (default=None)
        Determines where to truncate the bonding curve. The truncation point is given
        by D*truncate, where D is the total supply of tokens in the pool. Stableswap
        pools apply 0.0005 by default, and Cryptoswap pools apply 1 by default.

    resolution : int, optional (default=1000)
        The number of points to compute along the bonding curve.

    plot : bool, optional (default=False)
        Plots the bonding curves using Matplotlib.

    Returns
    -------
    pair_to_curve : dict
        Dictionary with coin index pairs as keys and lists of corresponding reserves
        as values. Each list of reserves is a list of pairs, where each pair consists
        of the reserves for the first and second coin of the corresponding pair.

    Example
    --------
    >>> import curvesim
    >>> pool_address = "0xbEbc44782C7dB0a1A60Cb6fe97d0b483032FF1C7"
    >>> pool = curvesim.pool.get(pool_address)
    >>> pair_to_curve = curvesim.bonding_curve(pool, plot=True)
    """
    if isinstance(pool, CurveMetaPool):
        combos: List[IndexPair] = [(0, 1)]
    else:
        combos = list(combinations(range(pool.n), 2))

    xp: List[int] = pool._xp()  # pylint: disable=protected-access

    if isinstance(pool, (CurveMetaPool, CurvePool, CurveRaiPool)):
        D: int = pool.D()
        if truncate is None:
            # This default value works for Stableswap, but will break Cryptoswap.
            # At this value, the graph usually cuts off cleanly around the points where
            # the Stableswap pool would depeg, as one stablecoin balance has reached
            # almost 100% of pool assets.

            truncate = 0.0005
    elif isinstance(pool, CurveCryptoPool):
        D = pool.D  # Don't recalcuate D - it will rebalance the bonding curve(s)
        if truncate is None:
            # 1 (int) "just works" for Cryptoswap. It extends the graph to the
            # point at which one coin, when valued at the price scale around which
            # liquidity is centered, is pushed to 100% of deposits `D` after starting
            # at (1 / pool.n) from the most recent rebalance. That should cover a
            # sufficient domain. The further away from (1 / pool.n) truncate is, the
            # more imbalanced the pool is at the ends of the graph.

            truncate = 1
    else:
        raise TypeError(f"Bonding curve calculation not supported for {type(pool)}")

    pair_to_curve: Dict[IndexPair, List[Point]] = {}
    current_points: Dict[IndexPair, Point] = {}
    for (i, j) in combos:
        truncated_D: int = int(D * truncate)
        x_limit: int = pool.get_y(j, i, truncated_D, xp)
        xs: List[int] = list(
            map(int, linspace(truncated_D, x_limit, resolution).round())
        )

        curve: List[Point] = []
        for x in xs:
            y: int = pool.get_y(i, j, x, xp)
            x_float, y_float = _denormalize((x, y), (i, j), pool)

            curve.append((x_float, y_float))

        pair_to_curve[(i, j)] = curve

        current_x: int = xp[i]
        current_y: int = xp[j]

        current_x_float, current_y_float = _denormalize(
            (current_x, current_y), (i, j), pool
        )

        current_points[(i, j)] = (current_x_float, current_y_float)

    if plot:
        labels: List[str] = pool.coin_names
        if not labels:
            labels = [f"Coin {str(label)}" for label in range(pool.n)]

        _plot_bonding_curve(pair_to_curve, current_points, labels)

    return pair_to_curve


def _denormalize(
    normalized_point: NormalizedPoint,
    index_pair: IndexPair,
    pool: Union[Stableswap, Cryptoswap],
) -> Point:
    """
    Converts a point made of integer balances (as if following EVM rules) to
    human-readable float form.
    """
    x, y = normalized_point
    i, j = index_pair

    assert i != j  # dev: x and y axes must use coins of different indices

    if isinstance(pool, (CurveMetaPool, CurvePool, CurveRaiPool)):
        x_factor: int = D_UNIT
        y_factor: int = D_UNIT
    elif isinstance(pool, CurveCryptoPool):
        x_factor = D_UNIT if i == 0 else pool.price_scale[i - 1]
        y_factor = D_UNIT if j == 0 else pool.price_scale[j - 1]

    x_float: float = x / x_factor
    y_float: float = y / y_factor
    point: Point = (x_float, y_float)

    return point


def _plot_bonding_curve(
    pair_to_curve: Dict[IndexPair, List[Point]],
    current_points: Dict[IndexPair, Point],
    labels: List[str],
) -> None:
    n: int = len(pair_to_curve)
    _, axs = plt.subplots(1, n, constrained_layout=True)
    if n == 1:
        axs = [axs]

    for pair, ax in zip(pair_to_curve, axs):
        curve: List[Point] = pair_to_curve[pair]
        xs, ys = zip(*curve)
        ax.plot(xs, ys, color="black")  # the entire bonding curve

        i, j = pair
        x, y = current_points[(i, j)]
        ax.scatter(x, y, s=40, color="black")  # A single dot at the current point
        ax.set_xlabel(labels[i])
        ax.set_ylabel(labels[j])

    plt.show()



"""
Utility functions for working with EVM-like account addresses.

This module provides custom versions of common address operations to ensure
standard handling of EVM addresses.
"""
__all__ = ["to_address", "is_address", "Address"]

from typing import NewType

from eth_typing import ChecksumAddress
from eth_utils import is_checksum_address, is_normalized_address, to_checksum_address

Address = NewType("Address", str)


def to_address(address_string: str) -> Address:
    """
    Convert given string to an EVM-like checksum address.

    Parameters
    ----------
    address_string: str
        string to convert to checksum address.

    Raises
    ------
    ValueError
        string cannot be coerced to a checksum address
    """
    checksum_address: ChecksumAddress = to_checksum_address(address_string)
    return Address(checksum_address)


def is_address(address_string: str, checksum=True) -> bool:
    """
    Check if given string is properly formatted as an
    EVM-like address, optionally as a checksum address.

    Parameters
    ----------
    address_string: str
        string to check for proper address formatting
    checksum: bool, default=True
        verify it is properly checksummed
    """
    if checksum:  # pylint: disable=no-else-return
        return is_checksum_address(address_string)
    else:
        return is_normalized_address(address_string)



from pprint import pformat

from numpy import isnan
from scipy.optimize import least_squares

from curvesim.logging import get_logger
from curvesim.templates.trader import Trade, Trader

from ..common import get_arb_trades

logger = get_logger(__name__)


class VolumeLimitedArbitrageur(Trader):
    """
    Computes, executes, and reports out arbitrage trades.
    """

    def compute_trades(self, prices, volume_limits):  # pylint: disable=arguments-differ
        """
        Computes trades to optimally arbitrage the pool, constrained by volume limits.

        Parameters
        ----------
        prices : dict
            Current market prices from the price_sampler.

        volume_limits : dict
            Current volume limits for each trading pair.


        Returns
        -------
        trades : list of :class:`Trade` objects
            List of trades to perform.

        additional_data: dict
            Dict of additional data to be passed to the state log as part of trade_data.
        """

        trades, errors, _ = multipair_optimal_arbitrage(
            self.pool, prices, volume_limits
        )
        return trades, {"price_errors": errors}


def multipair_optimal_arbitrage(  # noqa: C901  pylint: disable=too-many-locals
    pool, prices, limits
):
    """
    Computes trades to optimally arbitrage the pool, constrained by volume limits.

    Parameters
    ----------
    pool :
        Simulation interface to a subclass of :class:`Pool`.

    prices : dict
        Current market prices from the price_sampler.

    volume_limits : dict
        Current volume limits for each trading pair.

    Returns
    -------
    trades : List[Tuple]
        List of trades to perform.
        Trades are formatted as (coin_i, coin_j, trade_size)

    errors : List[Float]
        Post-trade price error between pool price and market price for each token pair.

    res : scipy.optimize.OptimizeResult
        Results object from the numerical optimizer.
    """
    all_trades = get_arb_trades(pool, prices)
    input_trades, skipped_trades = _apply_volume_limits(all_trades, limits, pool)

    if not input_trades:
        price_errors = _make_price_errors(skipped_trades=skipped_trades, pool=pool)
        return [], price_errors, None

    input_trades = _sort_trades_by_size(input_trades)
    least_squares_inputs = _make_least_squares_inputs(input_trades, limits)

    def post_trade_price_error_multi(amounts_in, price_targets, coin_pairs):
        with pool.use_snapshot_context():
            for coin_pair, amount_in in zip(coin_pairs, amounts_in):
                if isnan(amount_in):
                    dx = 0
                else:
                    dx = int(amount_in)

                min_size = pool.get_min_trade_size(coin_pair[0])
                if dx > min_size:
                    pool.trade(*coin_pair, dx)

            errors = []
            for coin_pair, price_target in zip(coin_pairs, price_targets):
                price = pool.price(*coin_pair, use_fee=True)
                errors.append(price - price_target)

        return errors

    # Find trades that minimize difference between
    # pool price and external market price
    trades = []
    try:
        res = least_squares(
            post_trade_price_error_multi,
            **least_squares_inputs,
            gtol=10**-15,
            xtol=10**-15,
        )

        # Record optimized trades
        for trade, amount_in in zip(input_trades, res.x):
            if isnan(amount_in):
                continue

            amount_in = int(amount_in)
            min_size = pool.get_min_trade_size(trade.coin_in)
            if amount_in > min_size:
                trades.append(Trade(trade.coin_in, trade.coin_out, amount_in))

        price_errors = _make_price_errors(input_trades, res.fun, skipped_trades, pool)

    except Exception:
        logger.error("Opt Arbs:\n %s", pformat(least_squares_inputs), exc_info=True)
        price_errors = _make_price_errors(skipped_trades=all_trades, pool=pool)
        res = None

    return trades, price_errors, res


def _apply_volume_limits(arb_trades, limits, pool):
    """
    Returns list of ArbTrades with amount_in set to min(limit, amount_in). Any trades
    limited to less than the pool's minimum trade size are excluded.
    """

    limited_arb_trades = []
    excluded_trades = []
    for trade in arb_trades:
        pair = trade.coin_in, trade.coin_out
        limited_amount_in = min(limits[pair], trade.amount_in)
        lim_trade = trade.replace_amount_in(limited_amount_in)

        if lim_trade.amount_in > pool.get_min_trade_size(lim_trade.coin_in):
            limited_arb_trades.append(lim_trade)
        else:
            excluded_trades.append(lim_trade)

    return limited_arb_trades, excluded_trades


def _sort_trades_by_size(trades):
    """Sorts trades by amount_in."""
    sorted_trades = sorted(trades, reverse=True, key=lambda t: t.amount_in)
    return sorted_trades


def _make_least_squares_inputs(trades, limits):
    """
    Returns a dict of trades, bounds, and targets formatted as kwargs for least_squares.
    """

    coins_in, coins_out, amounts_in, price_targets = zip(*trades)
    coin_pairs = tuple(zip(coins_in, coins_out))

    low_bound = [0] * len(trades)
    high_bound = [limits[pair] + 1 for pair in coin_pairs]

    return {
        "x0": amounts_in,
        "kwargs": {"price_targets": price_targets, "coin_pairs": coin_pairs},
        "bounds": (low_bound, high_bound),
    }


def _make_price_errors(trades=None, trade_errors=None, skipped_trades=None, pool=None):
    """
    Returns a dict mapping coin pairs to price errors.

    Parameters
    ----------
    trades :
        Trades input into the least_squares optimizer

    trade_errors :
        Price errors returned by the optimizer

    skipped_trades :
        Trades excluded from optimization

    pool :
        The pool used to compute the pool price for skipped trades

    Returns
    -------
    price_errors: Dict
        Maps coin pairs (tuple) to price_errors
    """
    price_errors = {}
    if trades:
        for trade, price_error in zip(trades, trade_errors):
            coin_pair = trade.coin_in, trade.coin_out
            price_errors[coin_pair] = price_error / trade.price_target

    if skipped_trades:
        for trade in skipped_trades:
            coin_pair = trade.coin_in, trade.coin_out
            price_error = pool.price(*coin_pair, use_fee=True) - trade.price_target
            price_errors[coin_pair] = price_error / trade.price_target

    return price_errors



"""
Contains variables and functions common to the arbitrage pipelines.
"""

from scipy.optimize import root_scalar

from curvesim.logging import get_logger
from curvesim.metrics import metrics as Metrics
from curvesim.templates.trader import ArbTrade

logger = get_logger(__name__)
DEFAULT_METRICS = [
    Metrics.Timestamp,
    Metrics.PoolValue,
    Metrics.PoolBalance,
    Metrics.PriceDepth,
    Metrics.PoolVolume,
    Metrics.ArbMetrics,
]


def get_arb_trades(pool, prices):
    """
    Returns triples of "trades", one for each coin pair in `combo`.

    Each trade is a triple consisting of size, ordered coin-pair,
    and price target to move the pool price to.

    Parameters
    ----------
    pool: SimPool
        Pool to arbitrage on

    prices : iterable
        External market prices for each coin-pair


    Returns
    -------
    trades: List[Tuple]
        List of triples (size, coins, price_target)
        "size": trade size
        "coins": in token, out token
        "price_target": price target for arbing the token pair
    """

    def post_trade_price_error(dx, coin_in, coin_out, price_target):
        with pool.use_snapshot_context():
            dx = int(dx)
            if dx > 0:
                pool.trade(coin_in, coin_out, dx)
            price = pool.price(coin_in, coin_out, use_fee=True)

        return price - price_target

    trades = []

    for pair in prices:
        coin_in, coin_out, target_price = _get_arb_direction(pair, pool, prices[pair])

        lower_bound = pool.get_min_trade_size(coin_in)
        profit_per_unit = post_trade_price_error(
            lower_bound, coin_in, coin_out, target_price
        )
        if profit_per_unit <= 0:
            trades.append(ArbTrade(coin_in, coin_out, 0, target_price))
            continue

        upper_bound = pool.get_max_trade_size(coin_in, coin_out)
        try:
            res = root_scalar(
                post_trade_price_error,
                args=(coin_in, coin_out, target_price),
                bracket=(lower_bound, upper_bound),
                method="brentq",
            )
            size = int(res.root)
        except ValueError:
            pool_price = pool.price(coin_in, coin_out)
            logger.error(
                "Opt_arb error: Pair: (%s, %s), Pool price: %s,"
                "Target Price: %s, Diff: %s",
                coin_in,
                coin_out,
                pool_price,
                target_price,
                pool_price - target_price,
            )
            size = 0

        trades.append(ArbTrade(coin_in, coin_out, size, target_price))

    return trades


def _get_arb_direction(pair, pool, market_price):
    i, j = pair
    price_error_i = pool.price(i, j) - market_price
    price_error_j = pool.price(j, i) - 1 / market_price

    if price_error_i >= price_error_j:
        target_price = market_price
        coin_in, coin_out = i, j
    else:
        target_price = 1 / market_price
        coin_in, coin_out = j, i

    return coin_in, coin_out, target_price
